# -*- coding: utf-8 -*-

import logging
import numpy as np
import leibniz as lbnz

from torch.nn.functional import conv3d, pad
from leibniz import cast

logger = logging.getLogger("leibniz")

default_device = -1


def get_device():
    return default_device


def set_device(ix):
    global default_device

    logger.info("reset device from %d to %d", default_device, ix)

    if ix != default_device:
        global centralx, centraly, centralz
        centralx = centralx.cuda(device=ix)
        centraly = centraly.cuda(device=ix)
        centralz = centralz.cuda(device=ix)

        global central5x, central5y, central5z
        central5x = central5x.cuda(device=ix)
        central5y = central5y.cuda(device=ix)
        central5z = central5z.cuda(device=ix)

        global sobel3x, sobel3y, sobel3z
        sobel3x = sobel3x.cuda(device=ix)
        sobel3y = sobel3y.cuda(device=ix)
        sobel3z = sobel3z.cuda(device=ix)

        global sobel5x, sobel5y, sobel5z
        sobel5x = sobel5x.cuda(device=ix)
        sobel5y = sobel5y.cuda(device=ix)
        sobel5z = sobel5z.cuda(device=ix)

        global sharr3x, sharr3y, sharr3z
        sobel3x = sobel3x.cuda(device=ix)
        sharr3y = sharr3y.cuda(device=ix)
        sharr3z = sharr3z.cuda(device=ix)

        global upwind_p2x, upwind_p2y, upwind_p2z
        upwind_p2x = upwind_p2x.cuda(device=ix)
        upwind_p2y = upwind_p2y.cuda(device=ix)
        upwind_p2z = upwind_p2z.cuda(device=ix)

        global upwind_m2x, upwind_m2y, upwind_m2z
        upwind_m2x = upwind_m2x.cuda(device=ix)
        upwind_m2y = upwind_m2y.cuda(device=ix)
        upwind_m2z = upwind_m2z.cuda(device=ix)

        global upwind_p3x, upwind_p3y, upwind_p3z
        upwind_p3x = upwind_p3x.cuda(device=ix)
        upwind_p3y = upwind_p3y.cuda(device=ix)
        upwind_p3z = upwind_p3z.cuda(device=ix)

        global upwind_m3x, upwind_m3y, upwind_m3z
        upwind_m3x = upwind_m3x.cuda(device=ix)
        upwind_m3y = upwind_m3y.cuda(device=ix)
        upwind_m3z = upwind_m3z.cuda(device=ix)

        global upwind_p4x, upwind_p4y, upwind_p4z
        upwind_p4x = upwind_p4x.cuda(device=ix)
        upwind_p4y = upwind_p4y.cuda(device=ix)
        upwind_p4z = upwind_p4z.cuda(device=ix)

        global upwind_m4x, upwind_m4y, upwind_m4z
        upwind_m4x = upwind_m4x.cuda(device=ix)
        upwind_m4y = upwind_m4y.cuda(device=ix)
        upwind_m4z = upwind_m4z.cuda(device=ix)

    default_device = ix


def fill_boundary_central(f):
    # Filling a variable to maintain 2nd-order accuracy at boundary points,
    # while compatible with conv3d using central-diff.
    f = pad(f, (1, 1, 1, 1, 1, 1), mode="replicate")  # 3D zero-padding
    f[:, :, 0, :, :] = 3 * f[:, :, 1, :, :] - 3 * f[:, :, 2, :, :] + f[:, :, 3, :, :]
    f[:, :, -1, :, :] = (
        3 * f[:, :, -2, :, :] - 3 * f[:, :, -3, :, :] + f[:, :, -4, :, :]
    )
    f[:, :, :, 0, :] = 3 * f[:, :, :, 1, :] - 3 * f[:, :, :, 2, :] + f[:, :, :, 3, :]
    f[:, :, :, -1, :] = (
        3 * f[:, :, :, -2, :] - 3 * f[:, :, :, -3, :] + f[:, :, :, -4, :]
    )
    f[:, :, :, :, 0] = 3 * f[:, :, :, :, 1] - 3 * f[:, :, :, :, 2] + f[:, :, :, :, 3]
    f[:, :, :, :, -1] = (
        3 * f[:, :, :, :, -2] - 3 * f[:, :, :, :, -3] + f[:, :, :, :, -4]
    )

    return f


centralx = (
    cast(
        [
            [
                [
                    [[0, 0, 0], [0, -1, 0], [0, 0, 0]],
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[0, 0, 0], [0, 1, 0], [0, 0, 0]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2
)
centraly = (
    cast(
        [
            [
                [
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[0, -1, 0], [0, 0, 0], [0, 1, 0]],
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2
)
centralz = (
    cast(
        [
            [
                [
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[0, 0, 0], [-1, 0, 1], [0, 0, 0]],
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2
)


def central(f):
    f = fill_boundary_central(f)

    return (
        conv3d(f, centralx),
        conv3d(f, centraly),
        conv3d(f, centralz),
    )


sobel3x = (
    -cast(
        [
            [
                [
                    [[+1, +2, +1], [+2, +4, +2], [+1, +2, +1]],
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[-1, -2, -1], [-2, -4, -2], [-1, -2, -1]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 32
)
sobel3y = (
    -cast(
        [
            [
                [
                    [[+1, +2, +1], [0, 0, 0], [-1, -2, -1]],
                    [[+2, +4, +2], [0, 0, 0], [-2, -4, -2]],
                    [[+1, +2, +1], [0, 0, 0], [-1, -2, -1]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 32
)
sobel3z = (
    -cast(
        [
            [
                [
                    [[+1, 0, -1], [+2, 0, -2], [+1, 0, -1]],
                    [[+2, 0, -2], [+4, 0, -4], [+2, 0, -2]],
                    [[+1, 0, -1], [+2, 0, -2], [+1, 0, -1]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 32
)


def sobel3(f):
    f = fill_boundary_central(f)

    return (
        conv3d(f, sobel3x),
        conv3d(f, sobel3y),
        conv3d(f, sobel3z),
    )


sharr3x = (
    -cast(
        [
            [
                [
                    [[+47, +162, +47], [+94, +324, +94], [+47, +162, +47]],
                    [[0, 0, 0], [0, 0, 0], [0, 0, 0]],
                    [[-47, -162, -47], [-94, -324, -94], [-47, -162, -47]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2048
)
sharr3y = (
    -cast(
        [
            [
                [
                    [[+47, +94, +47], [0, 0, 0], [-47, -94, -47]],
                    [[+162, +324, +162], [0, 0, 0], [-162, -324, -162]],
                    [[+47, +94, +47], [0, 0, 0], [-47, -94, -47]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2048
)
sharr3z = (
    -cast(
        [
            [
                [
                    [[+47, 0, -47], [+94, 0, -94], [+47, 0, -47]],
                    [[+162, 0, -162], [+324, 0, -324], [+162, 0, -162]],
                    [[+47, 0, -47], [+94, 0, -94], [+47, 0, -47]],
                ]
            ]
        ],
        device=lbnz.get_device(),
    )
    / 2048
)


def sharr3(f):
    f = fill_boundary_central(f)

    return (
        conv3d(f, sharr3x),
        conv3d(f, sharr3y),
        conv3d(f, sharr3z),
    )


def get_five_point_central_kernels():
    # 4th-order, kernel size = 5x5x5; diff_f = (-f_(i+2) + 8*f_(i+1) - 8*f_(i-1) + f_(i-2)) / 12
    five_point_diff = np.array([1, -8, 0, 8, -1]) / 12  # 5-point differential
    kernelx = np.zeros((1, 1, 5, 5, 5))
    kernelx[0, 0, :, 2, 2] = five_point_diff
    kernely = np.swapaxes(kernelx, 2, 3)
    kernelz = np.swapaxes(kernelx, 2, 4)

    return (
        cast(kernelx, device=lbnz.get_device()),
        cast(kernely, device=lbnz.get_device()),
        cast(kernelz, device=lbnz.get_device()),
    )


central5x, central5y, central5z = get_five_point_central_kernels()


def central5(f):
    f = fill_boundary_central(fill_boundary_central(f))

    return (conv3d(f, central5x), conv3d(f, central5y), conv3d(f, central5z))


def get_five_point_sobel_kernels():
    # get 5x5 sobel-way built kernels
    smooth_5_point = np.array([[1, 4, 6, 4, 1]])  # 1-D 5-point smoothing
    smooth_5x5 = (np.transpose(smooth_5_point) * smooth_5_point).reshape(
        (1, 5, 5)
    )  # 2-D 5x5 smoothing
    five_point_diff = (np.array([1, -8, 0, 8, -1]) / 12).reshape(
        5, 1, 1
    )  # 5-point differential
    # add the denominator to normalize the kernel parameters
    sobel_5x5_x = (smooth_5x5 * five_point_diff / 256).reshape(1, 1, 5, 5, 5)
    sobel_5x5_y = np.swapaxes(sobel_5x5_x, 2, 3)
    sobel_5x5_z = np.swapaxes(sobel_5x5_x, 2, 4)

    return (
        cast(sobel_5x5_x, device=lbnz.get_device()),
        cast(sobel_5x5_y, device=lbnz.get_device()),
        cast(sobel_5x5_z, device=lbnz.get_device()),
    )


sobel5x, sobel5y, sobel5z = get_five_point_sobel_kernels()


def sobel5(f):

    # this filling way can somehow bring errors, remains to be revised.
    f = fill_boundary_central(fill_boundary_central(f))

    return (
        conv3d(f, sobel5x),
        conv3d(f, sobel5y),
        conv3d(f, sobel5z),
    )


def fill_boundary_upwind(f, direction):
    if direction == "p":
        f = pad(f, (0, 1, 0, 1, 0, 1))  # one-side zero-padding
        f[:, :, -1, :, :] = 2 * f[:, :, -2, :, :] - f[:, :, -3, :, :]
        f[:, :, :, -1, :] = 2 * f[:, :, :, -2, :] - f[:, :, :, -3, :]
        f[:, :, :, :, -1] = 2 * f[:, :, :, :, -2] - f[:, :, :, :, -3]
    elif direction == "m":
        f = pad(f, (1, 0, 1, 0, 1, 0))  # one-side zero-padding
        f[:, :, 0, :, :] = 2 * f[:, :, 1, :, :] - f[:, :, 2, :, :]
        f[:, :, :, 0, :] = 2 * f[:, :, :, 1, :] - f[:, :, :, 2, :]
        f[:, :, :, :, 0] = 2 * f[:, :, :, :, 1] - f[:, :, :, :, 2]

    return f


def get_two_point_upwind_kernels(direction):
    # 'p' denotes 'plus', i.e. diff_u = (u_i+1) - u_i
    # 'm' denotes 'minus', i.e. diff_u = u_i - (u_i-1)

    upwind_x = np.zeros((1, 1, 2, 2, 2))
    upwind_diff = np.array([-1, 1])

    if direction == "p":
        upwind_x[0, 0, :, 0, 0] = upwind_diff
    elif direction == "m":
        upwind_x[0, 0, :, 1, 1] = upwind_diff

    upwind_y = np.swapaxes(upwind_x, 2, 3)
    upwind_z = np.swapaxes(upwind_x, 2, 4)

    return cast(upwind_x), cast(upwind_y), cast(upwind_z)


upwind_p2x, upwind_p2y, upwind_p2z = get_two_point_upwind_kernels("p")
upwind_m2x, upwind_m2y, upwind_m2z = get_two_point_upwind_kernels("m")


def upwind_p2(f):
    f = fill_boundary_upwind(f, "p")
    return (
        conv3d(f, upwind_p2x),
        conv3d(f, upwind_p2y),
        conv3d(f, upwind_p2z),
    )


def upwind_m2(f):
    f = fill_boundary_upwind(f, "m")

    return (
        conv3d(f, upwind_m2x),
        conv3d(f, upwind_m2y),
        conv3d(f, upwind_m2z),
    )


def get_three_point_upwind_kernels(direction):
    # 'p' denotes 'plus', i.e. diff_u = -(u_i+2) + 4 * (u_i+1) - 3 * u_i
    # 'm' denotes 'minus', i.e. diff_u = 3 * u_i - 4 * (u_i-1) + (u_i-2)

    upwind_x = np.zeros((1, 1, 3, 3, 3))

    if direction == "p":
        upwind_diff = np.array([-3, 4, -1]) / 2
        upwind_x[0, 0, :, 0, 0] = upwind_diff
    elif direction == "m":
        upwind_diff = np.array([1, -4, 3]) / 2
        upwind_x[0, 0, :, -1, -1] = upwind_diff

    upwind_y = np.swapaxes(upwind_x, 2, 3)
    upwind_z = np.swapaxes(upwind_x, 2, 4)

    return (
        cast(upwind_x, device=lbnz.get_device()),
        cast(upwind_y, device=lbnz.get_device()),
        cast(upwind_z, device=lbnz.get_device()),
    )


upwind_p3x, upwind_p3y, upwind_p3z = get_three_point_upwind_kernels("p")
upwind_m3x, upwind_m3y, upwind_m3z = get_three_point_upwind_kernels("m")


def upwind_p3(f):

    # this filling way can somehow bring errors, remains to be revised.
    f = fill_boundary_upwind(fill_boundary_upwind(f, "p"), "p")

    return (
        conv3d(f, upwind_p3x),
        conv3d(f, upwind_p3y),
        conv3d(f, upwind_p3z),
    )


def upwind_m3(f):

    # this filling way can somehow bring errors, remains to be revised.
    f = fill_boundary_upwind(fill_boundary_upwind(f, "m"), "m")

    return (
        conv3d(f, upwind_m3x),
        conv3d(f, upwind_m3y),
        conv3d(f, upwind_m3z),
    )


def get_four_point_upwind_kernels(direction):
    # 'p' denotes 'plus', i.e. diff_u = -(u_i+2) + 4 * (u_i+1) - 3 * u_i
    # 'm' denotes 'minus', i.e. diff_u = 3 * u_i - 4 * (u_i-1) + (u_i-2)

    upwind_x = np.zeros((1, 1, 4, 4, 4))

    if direction == "p":
        upwind_diff = np.array([-2, -3, 6, -1]) / 6
        upwind_x[0, 0, :, 1, 1] = upwind_diff
    elif direction == "m":
        upwind_diff = np.array([1, -6, 3, 2]) / 6
        upwind_x[0, 0, :, -2, -2] = upwind_diff

    upwind_y = np.swapaxes(upwind_x, 2, 3)
    upwind_z = np.swapaxes(upwind_x, 2, 4)

    return (
        cast(upwind_x, device=lbnz.get_device()),
        cast(upwind_y, device=lbnz.get_device()),
        cast(upwind_z, device=lbnz.get_device()),
    )


upwind_p4x, upwind_p4y, upwind_p4z = get_four_point_upwind_kernels("p")
upwind_m4x, upwind_m4y, upwind_m4z = get_four_point_upwind_kernels("m")


def upwind_p4(f):

    # this filling way can somehow bring errors, remains to be revised.
    f = fill_boundary_upwind(
        fill_boundary_upwind(fill_boundary_upwind(f, "m"), "p"), "p"
    )

    return (
        conv3d(f, upwind_p4x),
        conv3d(f, upwind_p4y),
        conv3d(f, upwind_p4z),
    )


def upwind_m4(f):

    # this filling way can somehow bring errors, remains to be revised.
    f = fill_boundary_upwind(
        fill_boundary_upwind(fill_boundary_upwind(f, "p"), "m"), "m"
    )

    return (
        conv3d(f, upwind_m4x),
        conv3d(f, upwind_m4y),
        conv3d(f, upwind_m4z),
    )



# -*- coding: utf-8 -*-

import unittest

import numpy as np

import leibniz as lbnz

from leibniz.core3d.vec3 import norm, cross, sub


class TestV3(unittest.TestCase):

    def setUp(self):
        self.zero = lbnz.cast(np.zeros([1, 1, 2, 2, 2]))
        self.one = lbnz.cast(np.ones([1, 1, 2, 2, 2]))

    def tearDown(self):
        pass

    def test_cross(self):
        i = self.one, self.zero, self.zero
        j = self.zero, self.one, self.zero
        k = self.zero, self.zero, self.one

        z = norm(sub(cross(i, j), k)).cpu().numpy()
        self.assertAlmostEqual(0, np.max(z))
        self.assertAlmostEqual(0, np.min(z))

        z = norm(sub(cross(j, k), i)).cpu().numpy()
        self.assertAlmostEqual(0, np.max(z))
        self.assertAlmostEqual(0, np.min(z))

        z = norm(sub(cross(k, i), j)).cpu().numpy()
        self.assertAlmostEqual(0, np.max(z))
        self.assertAlmostEqual(0, np.min(z))



# -*- coding: utf-8 -*-


_name_ = "x,y"
_params_ = ("x", "y")


def transform(x, y, **kwargs):
    return x, y


def dtransform(x, y, dx, dy, **kwargs):
    return dx, dy


def inverse(x, y, **kwargs):
    return x, y


def dinverse(x, y, dx, dy, **kwargs):
    return dx, dy



# -*- coding: utf-8 -*-

import unittest
import torch as th

from leibniz.nn.net import resnet
from leibniz.nn.layer.hyperbolic import HyperBasic
from leibniz.nn.layer.hyperbolic import HyperBottleneck


class TestResNet(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1D(self):
        net = resnet(1, 1, spatial=(32,))
        net(th.rand(1, 1, 32))
        net = resnet(1, 1, spatial=(32,), normalizor="instance")
        net(th.rand(1, 1, 32))
        net = resnet(1, 1, spatial=(32,), normalizor="layer")
        net(th.rand(1, 1, 32))
        net(th.rand(2, 1, 32))
        net = resnet(2, 1, spatial=(32,), ratio=0)
        net(th.rand(1, 2, 32))
        net(th.rand(2, 2, 32))

    def test2D(self):
        resnet(1, 1, spatial=(16, 16))
        resnet(1, 1, spatial=(16, 32))
        resnet(1, 1, spatial=(32, 16))
        net = resnet(1, 1, spatial=(32, 16))
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), normalizor="instance")
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), normalizor="layer")
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 64))
        net(th.rand(1, 1, 32, 64))
        net = resnet(1, 1, spatial=(16, 64))
        net(th.rand(1, 1, 32, 64))

    def test3D(self):
        resnet(1, 1, spatial=(16, 16, 16))
        resnet(1, 1, spatial=(32, 16, 16))
        resnet(1, 1, spatial=(16, 32, 16))
        resnet(1, 1, spatial=(16, 16, 32))
        resnet(1, 1, spatial=(11, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32))
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32), normalizor="instance")
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32), normalizor="layer")
        net(th.rand(1, 1, 4, 16, 32))
        net(th.rand(2, 1, 4, 16, 32))

    def testHyp1D(self):
        net = resnet(1, 1, spatial=(32,), block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resnet(1, 1, spatial=(32,), normalizor="instance", block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resnet(1, 1, spatial=(32,), normalizor="layer", block=HyperBasic)
        net(th.rand(1, 1, 16))
        net = resnet(1, 1, spatial=(32,), block=HyperBottleneck)
        net(th.rand(1, 1, 16))
        net = resnet(1, 1, spatial=(32,), normalizor="instance", block=HyperBottleneck)
        net(th.rand(1, 1, 16))
        net = resnet(1, 1, spatial=(32,), normalizor="layer", block=HyperBottleneck)
        net(th.rand(1, 1, 16))
        net(th.rand(2, 1, 16))

    def testHyp2D(self):
        net = resnet(1, 1, spatial=(32, 16), block=HyperBasic)
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), normalizor="instance", block=HyperBasic)
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), normalizor="layer", block=HyperBasic)
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), block=HyperBottleneck)
        net(th.rand(1, 1, 32, 16))
        net = resnet(
            1, 1, spatial=(32, 16), normalizor="instance", block=HyperBottleneck
        )
        net(th.rand(1, 1, 32, 16))
        net = resnet(1, 1, spatial=(32, 16), normalizor="layer", block=HyperBottleneck)
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))

    def testHyp2D2(self):
        net = resnet(
            17,
            1,
            factors=[-1, -1, -1, -1],
            vblks=[1, 1, 1, 1, 1],
            spatial=(32, 16),
            block=HyperBottleneck,
        )
        net(th.rand(1, 17, 32, 16))
        net(th.rand(2, 17, 32, 16))
        net = resnet(
            119,
            1,
            layers=5,
            ratio=0,
            factors=[-2, -1, -1, -1, -1],
            vblks=[1, 1, 1, 1, 1],
            spatial=(32, 16),
            block=HyperBottleneck,
        )
        net(th.rand(1, 119, 32, 16))
        net(th.rand(2, 119, 32, 16))
        net = resnet(
            119,
            1,
            layers=6,
            ratio=0,
            factors=[-1, -1, -1, -1, -1, -1],
            vblks=[1, 1, 1, 1, 1, 1],
            spatial=(32, 16),
            block=HyperBasic,
        )
        net(th.rand(1, 119, 32, 16))
        net(th.rand(2, 119, 32, 16))

    def testHyp3D(self):
        net = resnet(1, 1, spatial=(4, 16, 32), block=HyperBasic)
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32), normalizor="instance", block=HyperBasic)
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32), normalizor="layer", block=HyperBasic)
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(1, 1, spatial=(4, 16, 32), block=HyperBottleneck)
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(
            1, 1, spatial=(4, 16, 32), normalizor="instance", block=HyperBottleneck
        )
        net(th.rand(1, 1, 4, 16, 32))
        net = resnet(
            1, 1, spatial=(4, 16, 32), normalizor="layer", block=HyperBottleneck
        )
        net(th.rand(1, 1, 4, 16, 32))
        net(th.rand(2, 1, 4, 16, 32))



# -*- coding: utf-8 -*-

import unittest
import torch as th

from leibniz.nn.net import resunetz


class TestResUNetZ(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test1D(self):
        net = resunetz(2, 1, spatial=(32,))
        net(th.rand(1, 2, 32))
        net = resunetz(1, 2, spatial=(32,), normalizor="relu6")
        net(th.rand(1, 1, 32))
        net = resunetz(1, 2, spatial=(32,), normalizor="tanh")
        net(th.rand(1, 1, 32))
        net = resunetz(1, 1, spatial=(32,), normalizor="softmax")
        net(th.rand(1, 1, 32))
        net = resunetz(1, 1, spatial=(32,), normalizor="sigmoid")
        net(th.rand(1, 1, 32))
        net(th.rand(2, 1, 32))
        net = resunetz(2, 1, spatial=(32,), ratio=0)
        net(th.rand(1, 2, 32))
        net(th.rand(2, 2, 32))

    def test2D(self):
        resunetz(1, 1, spatial=(16, 16))
        resunetz(1, 1, spatial=(16, 32))
        resunetz(1, 1, spatial=(32, 16))
        net = resunetz(1, 2, spatial=(32, 16))
        net(th.rand(1, 1, 32, 16))
        net = resunetz(2, 1, spatial=(32, 16), normalizor="relu6")
        net(th.rand(1, 2, 32, 16))
        net = resunetz(1, 1, spatial=(32, 16), normalizor="tanh")
        net(th.rand(1, 1, 32, 16))
        net = resunetz(1, 1, spatial=(32, 16), normalizor="softmax")
        net(th.rand(1, 1, 32, 16))
        net = resunetz(1, 1, spatial=(32, 16), normalizor="sigmoid")
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))
        net = resunetz(1, 1, spatial=(32, 64))
        net(th.rand(1, 1, 32, 64))
        net = resunetz(1, 1, spatial=(16, 64))
        net(th.rand(1, 1, 32, 64))

    def test3D(self):
        resunetz(1, 1, spatial=(16, 16, 16))
        resunetz(1, 1, spatial=(32, 16, 16))
        resunetz(1, 1, spatial=(16, 32, 16))
        resunetz(1, 1, spatial=(16, 16, 32))
        resunetz(1, 1, spatial=(11, 16, 32))
        net = resunetz(1, 1, spatial=(16, 16, 16))
        net(th.rand(2, 1, 16, 16, 16))
        net = resunetz(1, 1, spatial=(16, 16, 16), normalizor="relu6")
        net(th.rand(2, 1, 16, 16, 16))
        net = resunetz(1, 1, spatial=(16, 16, 16), normalizor="tanh")
        net(th.rand(2, 1, 16, 16, 16))
        net = resunetz(1, 1, spatial=(16, 16, 16), normalizor="softmax")
        net(th.rand(2, 1, 16, 16, 16))
        net = resunetz(1, 1, spatial=(16, 16, 16), normalizor="sigmoid")
        net(th.rand(2, 1, 16, 16, 16))
        net(th.rand(2, 1, 16, 16, 16))



# -*- coding: utf-8 -*-

import logging

import numpy as np
import torch as th
import torch.nn as nn

from leibniz.nn.conv import (
    DepthwiseSeparableConv1d,
    DepthwiseSeparableConv2d,
    DepthwiseSeparableConv3d,
)
from leibniz.nn.layer.cbam import CBAM
from leibniz.nn.net.hyptube import HypTube
from leibniz.nn.net.unet import Enconv, Deconv, Block, Transform

logger = logging.getLogger()
logger.setLevel(logging.INFO)


class UNet2(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        block=None,
        attn=CBAM,
        relu=None,
        layers=4,
        ratio=2,
        enhencer=None,
        ksize_in=7,
        dropout_prob=0.1,
        vblks=None,
        hblks=None,
        scales=None,
        factors=None,
        spatial=(256, 256),
        normalizor="batch",
        padding=None,
        final_normalized=True,
    ):
        super().__init__()

        extension = block.extension
        lrd = block.least_required_dim

        spatial = np.array(spatial, dtype=np.int32)
        dim = len(spatial)
        self.dim = dim
        Conv = self.get_conv_for_prepare()
        TConv = self.get_conv_for_transform()

        scales = np.array(scales)
        if scales.shape[0] != layers:
            raise ValueError("scales should have %d layers at dim 0!" % layers)
        if len(scales.shape) == 1:
            scales = scales.reshape([layers, 1])
        if len(scales.shape) != 2:
            raise ValueError(
                "scales should have length 2 to be compatible with spatial dimensions!"
            )

        ratio = np.exp2(ratio)
        factors = np.array(factors + [0.0])
        scales = np.exp2(scales)
        factors = np.exp2(factors)
        num_filters = int(in_channels * ratio)

        self.final_normalized = final_normalized
        self.ratio = ratio
        self.hblks = hblks
        self.vblks = vblks
        self.scales = scales
        self.factors = factors
        logger.info("---------------------------------------")
        logger.info("ratio: %f", ratio)
        logger.info("vblks: [%s]", ", ".join(map(str, vblks)))
        logger.info("hblks: [%s]", ", ".join(map(str, hblks)))
        logger.info("scales: [%s]", ", ".join(map(str, scales)))
        logger.info("factors: [%s]", ", ".join(map(str, factors[0:4])))
        logger.info("---------------------------------------")

        self.exceeded = np.any(np.cumprod(scales, axis=0) * spatial < 1) or np.any(
            (in_channels * ratio * np.cumprod(factors)) < lrd
        )
        if not self.exceeded:
            self.layers = layers
            self.in_channels = in_channels
            self.num_filters = num_filters
            self.out_channels = out_channels

            if enhencer is None:
                if self.dim == 2:
                    enhencer = HypTube
            self.enhencer_in = None
            self.enhencer_out = None
            self.enhencer_mid = None

            if relu is None:
                relu = nn.ReLU(inplace=True)

            if attn is None:
                attn = CBAM

            ex = extension
            c0 = int(ex * num_filters)
            if padding:
                self.conv_padding = 0
                self.iconv = nn.Sequential(
                    padding,
                    Conv(
                        in_channels,
                        c0,
                        kernel_size=ksize_in,
                        padding=(ksize_in - 1) // 2,
                        groups=1,
                    ),
                )
                self.oconv = nn.Sequential(
                    padding,
                    Conv(
                        c0,
                        out_channels,
                        kernel_size=3,
                        padding=self.conv_padding,
                        bias=False,
                        groups=1,
                    ),
                )
            else:
                self.conv_padding = 1
                self.iconv = Conv(
                    in_channels,
                    c0,
                    kernel_size=ksize_in,
                    padding=(ksize_in - 1) // 2,
                    groups=1,
                )
                self.oconv = Conv(
                    c0 * 2,
                    out_channels,
                    kernel_size=3,
                    padding=self.conv_padding,
                    bias=False,
                    groups=1,
                )

            if final_normalized:
                self.relu6 = nn.ReLU6()

            self.enconvs = nn.ModuleList()
            self.dnforms = nn.ModuleList()
            self.hzforms = nn.ModuleList()
            self.upforms = nn.ModuleList()
            self.deconvs = nn.ModuleList()
            self.mdconvs = nn.ModuleList()

            self.spatial = [np.array(spatial, dtype=np.int32)]
            self.channel_sizes = [np.array(c0, dtype=np.int32)]
            for ix in range(layers):
                least_factor = ex
                scale, factor = scales[ix], factors[ix]
                self.spatial.append(np.array(self.spatial[ix] * scale, dtype=np.int32))
                self.channel_sizes.append(
                    np.array(
                        self.channel_sizes[ix] * factor // least_factor * least_factor,
                        dtype=np.int32,
                    )
                )

                ci, co = (
                    self.channel_sizes[ix].item(),
                    self.channel_sizes[ix + 1].item(),
                )
                szi, szo = self.spatial[ix + 1], self.spatial[ix]
                logger.info("%d - ci: %d, co: %d", ix, ci, co)
                logger.info(
                    "%d - szi: [%s], szo: [%s]",
                    ix,
                    ", ".join(map(str, szi)),
                    ", ".join(map(str, szo)),
                )

                self.exceeded = (
                    self.exceeded
                    or ci < lrd
                    or co < lrd
                    or szi.min() < 1
                    or szo.min() < 1
                )
                if not self.exceeded:
                    try:
                        dropout_flag = (layers - ix) * 3 < layers
                        dropout = dropout_prob if dropout_flag else -1
                        self.mdconvs.append(
                            Block(
                                Deconv(co, ci, size=szo, conv=TConv, padding=padding),
                                activation=True,
                                dropout=False,
                                relu=relu,
                                attn=attn,
                                dim=self.dim,
                                normalizor=normalizor,
                                conv=TConv,
                            )
                        )
                        self.enconvs.append(
                            Block(
                                Enconv(ci, co, size=szi, conv=TConv, padding=padding),
                                activation=True,
                                dropout=dropout,
                                relu=relu,
                                attn=attn,
                                dim=self.dim,
                                normalizor=normalizor,
                                conv=TConv,
                            )
                        )
                        self.dnforms.append(
                            Transform(
                                co,
                                co,
                                nblks=vblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                        self.hzforms.append(
                            Transform(
                                co * 2,
                                co * 2,
                                nblks=hblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                        self.deconvs.append(
                            Block(
                                Deconv(
                                    co * 3, ci, size=szo, conv=TConv, padding=padding
                                ),
                                activation=True,
                                dropout=False,
                                relu=relu,
                                attn=attn,
                                dim=self.dim,
                                normalizor=normalizor,
                                conv=TConv,
                            )
                        )
                        self.upforms.append(
                            Transform(
                                ci,
                                ci,
                                nblks=vblks[ix],
                                block=block,
                                relu=relu,
                                conv=TConv,
                            )
                        )
                    except Exception as e:
                        logger.exception(e)
                        self.exceeded = True
                else:
                    logger.error("scales are exceeded!")
                    raise ValueError("scales exceeded!")

            if self.dim == 2 and enhencer is not None:
                self.enhencer_mid = enhencer(co, (c0 + 1) // 2, co)

    def get_conv_for_prepare(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def get_conv_for_transform(self):
        if self.dim == 1:
            conv = DepthwiseSeparableConv1d
        elif self.dim == 2:
            conv = DepthwiseSeparableConv2d
        elif self.dim == 3:
            conv = DepthwiseSeparableConv3d
        else:
            raise ValueError("dim %d is not supported!" % self.dim)
        return conv

    def forward(self, x):
        if self.exceeded:
            raise ValueError("scales exceeded!")

        dnt = self.iconv(x)

        dnts, encs, mids = [], [], []
        for ix in range(self.layers):
            dnt, enc = self.dnforms[ix](self.enconvs[ix](dnt))
            mid = self.mdconvs[ix](enc)
            dnts.append(dnt)
            encs.append(enc)
            mids.append(mid)

        if self.enhencer_mid is not None:
            upt = self.enhencer_mid(dnt)
        else:
            upt = dnt

        for ix in range(self.layers - 1, -1, -1):
            enc = encs[ix]
            if ix == self.layers - 1:
                mid = upt

            inp = th.cat([enc, mid], dim=1)
            trans = self.hzforms[ix]
            hzt, _ = trans(inp)
            upt, dec = self.upforms[ix](self.deconvs[ix](upt, hzt))
            mid = mids[ix]

        return self.oconv(th.cat([upt, mid], dim=1))



# -*- coding: utf-8 -*-

import torch as th


_name_ = "theta,phi,r"
_params_ = ("theta", "phi", "r")


def transform(theta, phi, r, **kwargs):
    x = r * th.cos(phi) * th.cos(theta)
    y = r * th.cos(phi) * th.sin(theta)
    z = r * th.sin(phi)
    return x, y, z


def dtransform(theta, phi, r, dtheta, dphi, dr, **kwargs):
    dx = (
        -r * th.sin(phi) * th.cos(theta) * dphi
        - r * th.cos(phi) * th.sin(theta) * dtheta
        + th.cos(phi) * th.cos(theta) * dr
    )
    dy = (
        -r * th.sin(phi) * th.sin(theta) * dphi
        + r * th.cos(phi) * th.cos(theta) * dtheta
        + th.cos(phi) * th.sin(theta) * dr
    )
    dz = r * th.cos(phi) * dphi + th.sin(phi) * dr

    return dx, dy, dz


def inverse(x, y, z, **kwargs):
    r = th.sqrt(x * x + y * y + z * z)
    theta = th.atan2(y, x)
    phi = th.asin(z / r)
    return theta, phi, r


def dinverse(x, y, z, dx, dy, dz, **kwargs):
    r = th.sqrt(x * x + y * y + z * z)

    dr = (x * dx + y * dy + z * dz) / r
    dtheta = (x * dy - y * dx) / (x * x + y * y)
    dphi = (dz - z / r * dr) / th.sqrt(x * x + y * y)

    return dtheta, dphi, dr



import os
import torch as th

import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt  # noqa: E402

from matplotlib.animation import FuncAnimation  # noqa: E402
from leibniz.diffeq import odeint as odeint  # noqa: E402


def binary(tensor):
    return th.where(tensor > lbnz.zero, lbnz.one, lbnz.zero)


def draw_3D_plot(x, y, fld, fig_name="test.png"):
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    fld = fld.cpu().numpy()

    fig = plt.figure()
    ax = fig.gca(projection="3d")
    c = ax.plot_surface(
        x,
        y,
        fld,
        rstride=1,
        cstride=1,
        cmap="binary",
        edgecolor="none",
        vmin=-2,
        vmax=2,
    )
    ax.set_xlim(1, 16)
    ax.set_ylim(1, 6)
    ax.set_zlim(-2, 2)

    fig.colorbar(c)
    plt.savefig(os.path.join("results", fig_name))
    plt.close()


def draw_animation(sequence, x, y, gif_name="test.gif", frames=51, mode="3d"):
    gif_name = os.path.join("results", gif_name)
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    sequence = [item.cpu().numpy() for item in sequence]

    fig = plt.figure()

    def update(n):
        print("drawing %2dth image in animation" % n)
        plt.clf()
        ax = fig.gca(projection="3d")
        c = ax.plot_surface(
            x,
            y,
            sequence[n],
            rstride=1,
            cstride=1,
            cmap="binary",
            edgecolor="none",
            vmin=-2,
            vmax=2,
        )
        ax.set_xlim(1, 16)
        ax.set_ylim(1, 6)
        ax.set_zlim(-2, 2)
        fig.colorbar(c)

    ani = FuncAnimation(
        fig,
        update,
        frames=frames,
        interval=1000.0 / 4,
        save_count=100,
        repeat_delay=1000,
        repeat=True,
    )

    # save it into gif (need imagemagick)
    ani.save(gif_name[:-4] + mode + ".gif", writer="imagemagick")
    plt.close()


if __name__ == "__main__":
    import leibniz as lbnz
    from leibniz.core3d.gridsys.regular3 import RegularGrid

    lbnz.bind(
        RegularGrid(
            basis="x,y,z",
            W=51,
            L=151,
            H=51,
            east=16.0,
            west=1.0,
            north=6.0,
            south=1.0,
            upper=6.0,
            lower=1.0,
        )
    )
    lbnz.use("x,y,z")

    # dt = 0.05
    # t_iter = 50
    fld = (
        binary((lbnz.x - 8) * (9 - lbnz.x))
        * binary((lbnz.y - 3) * (4 - lbnz.y))
        * binary((lbnz.z - 3) * (4 - lbnz.z))
    )

    wind = lbnz.one, lbnz.zero, lbnz.zero

    def derivitive(t, clouds):
        return -lbnz.upwind(wind, clouds)

    pred = odeint(derivitive, fld, th.arange(0, 7, 1 / 100), method="rk4")
    # sequence = [pred[i].squeeze()[..., lbnz.H // 2] for i in range(700) if i % 10 == 0]
    # draw_animation(sequence, lbnz.x.squeeze()[..., 0], lbnz.y.squeeze()[..., 0], 'simple_advection(upwind4).gif',
    #               mode='3d', frames=len(sequence))



# Based on https://github.com/tensorflow/tensorflow/tree/master/tensorflow/contrib/integrate
import torch
from .misc import (
    _scaled_dot_product,
    _convert_to_tensor,
    _is_finite,
    _select_initial_step,
    _handle_unused_kwargs,
    _is_iterable,
    _optimal_step_size,
    _compute_error_ratio,
)
from .solvers import AdaptiveStepsizeODESolver
from .interp import _interp_fit, _interp_evaluate
from .rk_common import _RungeKuttaState, _ButcherTableau, _runge_kutta_step

_DORMAND_PRINCE_SHAMPINE_TABLEAU = _ButcherTableau(
    alpha=[1 / 5, 3 / 10, 4 / 5, 8 / 9, 1.0, 1.0],
    beta=[
        [1 / 5],
        [3 / 40, 9 / 40],
        [44 / 45, -56 / 15, 32 / 9],
        [19372 / 6561, -25360 / 2187, 64448 / 6561, -212 / 729],
        [9017 / 3168, -355 / 33, 46732 / 5247, 49 / 176, -5103 / 18656],
        [35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84],
    ],
    c_sol=[35 / 384, 0, 500 / 1113, 125 / 192, -2187 / 6784, 11 / 84, 0],
    c_error=[
        35 / 384 - 1951 / 21600,
        0,
        500 / 1113 - 22642 / 50085,
        125 / 192 - 451 / 720,
        -2187 / 6784 - -12231 / 42400,
        11 / 84 - 649 / 6300,
        -1.0 / 60.0,
    ],
)

DPS_C_MID = [
    6025192743 / 30085553152 / 2,
    0,
    51252292925 / 65400821598 / 2,
    -2691868925 / 45128329728 / 2,
    187940372067 / 1594534317056 / 2,
    -1776094331 / 19743644256 / 2,
    11237099 / 235043384 / 2,
]


def _interp_fit_dopri5(y0, y1, k, dt, tableau=_DORMAND_PRINCE_SHAMPINE_TABLEAU):
    """Fit an interpolating polynomial to the results of a Runge-Kutta step."""
    dt = dt.type_as(y0[0])
    y_mid = tuple(
        y0_ + _scaled_dot_product(dt, DPS_C_MID, k_) for y0_, k_ in zip(y0, k)
    )
    f0 = tuple(k_[0] for k_ in k)
    f1 = tuple(k_[-1] for k_ in k)
    return _interp_fit(y0, y1, y_mid, f0, f1, dt)


def _abs_square(x):
    return torch.mul(x, x)


def _ta_append(list_of_tensors, value):
    """Append a value to the end of a list of PyTorch tensors."""
    list_of_tensors.append(value)
    return list_of_tensors


class Dopri5Solver(AdaptiveStepsizeODESolver):

    def __init__(
        self,
        func,
        y0,
        rtol,
        atol,
        first_step=None,
        safety=0.9,
        ifactor=10.0,
        dfactor=0.2,
        max_num_steps=2**31 - 1,
        **unused_kwargs
    ):
        _handle_unused_kwargs(self, unused_kwargs)
        del unused_kwargs

        self.func = func
        self.y0 = y0
        self.rtol = rtol if _is_iterable(rtol) else [rtol] * len(y0)
        self.atol = atol if _is_iterable(atol) else [atol] * len(y0)
        self.first_step = first_step
        self.safety = _convert_to_tensor(
            safety, dtype=torch.float64, device=y0[0].device
        )
        self.ifactor = _convert_to_tensor(
            ifactor, dtype=torch.float64, device=y0[0].device
        )
        self.dfactor = _convert_to_tensor(
            dfactor, dtype=torch.float64, device=y0[0].device
        )
        self.max_num_steps = _convert_to_tensor(
            max_num_steps, dtype=torch.int32, device=y0[0].device
        )

    def before_integrate(self, t):
        f0 = self.func(t[0].type_as(self.y0[0]), self.y0)
        if self.first_step is None:
            first_step = _select_initial_step(
                self.func, t[0], self.y0, 4, self.rtol[0], self.atol[0], f0=f0
            ).to(t)
        else:
            first_step = _convert_to_tensor(0.01, dtype=t.dtype, device=t.device)
        self.rk_state = _RungeKuttaState(
            self.y0, f0, t[0], t[0], first_step, interp_coeff=[self.y0] * 5
        )

    def advance(self, next_t):
        """Interpolate through the next time point, integrating as necessary."""
        n_steps = 0
        while next_t > self.rk_state.t1:
            assert (
                n_steps < self.max_num_steps
            ), "max_num_steps exceeded ({}>={})".format(n_steps, self.max_num_steps)
            self.rk_state = self._adaptive_dopri5_step(self.rk_state)
            n_steps += 1
        return _interp_evaluate(
            self.rk_state.interp_coeff, self.rk_state.t0, self.rk_state.t1, next_t
        )

    def _adaptive_dopri5_step(self, rk_state):
        """Take an adaptive Runge-Kutta step to integrate the ODE."""
        y0, f0, _, t0, dt, interp_coeff = rk_state
        ########################################################
        #                      Assertions                      #
        ########################################################
        assert t0 + dt > t0, "underflow in dt {}".format(dt.item())
        for y0_ in y0:
            assert _is_finite(
                torch.abs(y0_)
            ), "non-finite values in state `y`: {}".format(y0_)
        y1, f1, y1_error, k = _runge_kutta_step(
            self.func, y0, f0, t0, dt, tableau=_DORMAND_PRINCE_SHAMPINE_TABLEAU
        )

        ########################################################
        #                     Error Ratio                      #
        ########################################################
        mean_sq_error_ratio = _compute_error_ratio(
            y1_error, atol=self.atol, rtol=self.rtol, y0=y0, y1=y1
        )
        accept_step = (torch.tensor(mean_sq_error_ratio) <= 1).all()

        ########################################################
        #                   Update RK State                    #
        ########################################################
        y_next = y1 if accept_step else y0
        f_next = f1 if accept_step else f0
        t_next = t0 + dt if accept_step else t0
        interp_coeff = (
            _interp_fit_dopri5(y0, y1, k, dt) if accept_step else interp_coeff
        )
        dt_next = _optimal_step_size(
            dt,
            mean_sq_error_ratio,
            safety=self.safety,
            ifactor=self.ifactor,
            dfactor=self.dfactor,
            order=5,
        )
        rk_state = _RungeKuttaState(y_next, f_next, t0, t_next, dt_next, interp_coeff)
        return rk_state



import torch
import torch.nn as nn
from . import odeint
from .misc import _flatten, _flatten_convert_none_to_zeros


class OdeintAdjointMethod(torch.autograd.Function):

    @staticmethod
    def forward(ctx, *args):
        assert len(args) >= 8, "Internal error: all arguments required."
        y0, func, t, flat_params, rtol, atol, method, options = (
            args[:-7],
            args[-7],
            args[-6],
            args[-5],
            args[-4],
            args[-3],
            args[-2],
            args[-1],
        )

        ctx.func, ctx.rtol, ctx.atol, ctx.method, ctx.options = (
            func,
            rtol,
            atol,
            method,
            options,
        )

        with torch.no_grad():
            ans = odeint(
                func, y0, t, rtol=rtol, atol=atol, method=method, options=options
            )
        ctx.save_for_backward(t, flat_params, *ans)
        return ans

    @staticmethod
    def backward(ctx, *grad_output):

        t, flat_params, *ans = ctx.saved_tensors
        ans = tuple(ans)
        func, rtol, atol, method, options = (
            ctx.func,
            ctx.rtol,
            ctx.atol,
            ctx.method,
            ctx.options,
        )
        n_tensors = len(ans)
        f_params = tuple(func.parameters())

        # TODO: use a nn.Module and call odeint_adjoint to implement higher order derivatives.
        def augmented_dynamics(t, y_aug):
            # Dynamics of the original system augmented with
            # the adjoint wrt y, and an integrator wrt t and args.
            y, adj_y = (
                y_aug[:n_tensors],
                y_aug[n_tensors : 2 * n_tensors],
            )  # Ignore adj_time and adj_params.

            with torch.set_grad_enabled(True):
                t = t.to(y[0].device).detach().requires_grad_(True)
                y = tuple(y_.detach().requires_grad_(True) for y_ in y)
                func_eval = func(t, y)
                vjp_t, *vjp_y_and_params = torch.autograd.grad(
                    func_eval,
                    (t,) + y + f_params,
                    tuple(-adj_y_ for adj_y_ in adj_y),
                    allow_unused=True,
                    retain_graph=True,
                )
            vjp_y = vjp_y_and_params[:n_tensors]
            vjp_params = vjp_y_and_params[n_tensors:]

            # autograd.grad returns None if no gradient, set to zero.
            vjp_t = torch.zeros_like(t) if vjp_t is None else vjp_t
            vjp_y = tuple(
                torch.zeros_like(y_) if vjp_y_ is None else vjp_y_
                for vjp_y_, y_ in zip(vjp_y, y)
            )
            vjp_params = _flatten_convert_none_to_zeros(vjp_params, f_params)

            if len(f_params) == 0:
                vjp_params = torch.tensor(0.0).to(vjp_y[0])
            return (*func_eval, *vjp_y, vjp_t, vjp_params)

        T = ans[0].shape[0]
        with torch.no_grad():
            adj_y = tuple(grad_output_[-1] for grad_output_ in grad_output)
            adj_params = torch.zeros_like(flat_params)
            adj_time = torch.tensor(0.0).to(t)
            time_vjps = []
            for i in range(T - 1, 0, -1):

                ans_i = tuple(ans_[i] for ans_ in ans)
                grad_output_i = tuple(grad_output_[i] for grad_output_ in grad_output)
                func_i = func(t[i], ans_i)

                # Compute the effect of moving the current time measurement point.
                dLd_cur_t = sum(
                    torch.dot(func_i_.reshape(-1), grad_output_i_.reshape(-1)).reshape(
                        1
                    )
                    for func_i_, grad_output_i_ in zip(func_i, grad_output_i)
                )
                adj_time = adj_time - dLd_cur_t
                time_vjps.append(dLd_cur_t)

                # Run the augmented system backwards in time.
                if adj_params.numel() == 0:
                    adj_params = torch.tensor(0.0).to(adj_y[0])
                aug_y0 = (*ans_i, *adj_y, adj_time, adj_params)
                aug_ans = odeint(
                    augmented_dynamics,
                    aug_y0,
                    torch.tensor([t[i], t[i - 1]]),
                    rtol=rtol,
                    atol=atol,
                    method=method,
                    options=options,
                )

                # Unpack aug_ans.
                adj_y = aug_ans[n_tensors : 2 * n_tensors]
                adj_time = aug_ans[2 * n_tensors]
                adj_params = aug_ans[2 * n_tensors + 1]

                adj_y = tuple(
                    adj_y_[1] if len(adj_y_) > 0 else adj_y_ for adj_y_ in adj_y
                )
                if len(adj_time) > 0:
                    adj_time = adj_time[1]
                if len(adj_params) > 0:
                    adj_params = adj_params[1]

                adj_y = tuple(
                    adj_y_ + grad_output_[i - 1]
                    for adj_y_, grad_output_ in zip(adj_y, grad_output)
                )

                del aug_y0, aug_ans

            time_vjps.append(adj_time)
            time_vjps = torch.cat(time_vjps[::-1])

            return (*adj_y, None, time_vjps, adj_params, None, None, None, None, None)


def odeint_adjoint(func, y0, t, rtol=1e-6, atol=1e-12, method=None, options=None):

    # We need this in order to access the variables inside this module,
    # since we have no other way of getting variables along the execution path.
    if not isinstance(func, nn.Module):
        raise ValueError("func is required to be an instance of nn.Module.")

    tensor_input = False
    if torch.is_tensor(y0):

        class TupleFunc(nn.Module):

            def __init__(self, base_func):
                super(TupleFunc, self).__init__()
                self.base_func = base_func

            def forward(self, t, y):
                return (self.base_func(t, y[0]),)

        tensor_input = True
        y0 = (y0,)
        func = TupleFunc(func)

    flat_params = _flatten(func.parameters())
    ys = OdeintAdjointMethod.apply(
        *y0, func, t, flat_params, rtol, atol, method, options
    )

    if tensor_input:
        ys = ys[0]
    return ys



# -*- coding: utf-8 -*-

import unittest
import torch as th
import torch.nn as nn

from torch.nn import Tanh, MSELoss, Sequential
from leibniz.nn.activation import Atanh
from leibniz.nn.normalizor import PWLNormalizor


class TestNormalizor(unittest.TestCase):

    def setUp(self):
        self.tanh = Tanh()
        self.atanh = Atanh()
        self.pwln = PWLNormalizor(2, 16, mean=1.4, std=2.5)

    def tearDown(self):
        pass

    def func(self, x):
        return th.cat(
            [
                x[:, 0:1] * 2 + 0.5 * th.cos(x[:, 0:1] * 8),
                x[:, 1:2] * 3 - 0.5 * th.sin(x[:, 1:2] * 7),
            ],
            dim=1,
        )

    def testAtanh(self):
        self.assertAlmostEqual(
            6.0, self.atanh(self.tanh(th.as_tensor(6.0))).numpy(), places=3
        )
        self.assertAlmostEqual(
            5.0, self.atanh(self.tanh(th.as_tensor(5.0))).numpy(), places=3
        )
        self.assertAlmostEqual(
            4.0, self.atanh(self.tanh(th.as_tensor(4.0))).numpy(), places=4
        )
        self.assertAlmostEqual(
            3.0, self.atanh(self.tanh(th.as_tensor(3.0))).numpy(), places=5
        )
        self.assertAlmostEqual(2.0, self.atanh(self.tanh(th.as_tensor(2.0))).numpy())
        self.assertAlmostEqual(1.0, self.atanh(self.tanh(th.as_tensor(1.0))).numpy())

        self.assertAlmostEqual(
            -1.0, self.atanh(self.tanh(th.as_tensor(-1.0))).numpy(), places=6
        )
        self.assertAlmostEqual(
            -2.0, self.atanh(self.tanh(th.as_tensor(-2.0))).numpy(), places=5
        )
        self.assertAlmostEqual(
            -3.0, self.atanh(self.tanh(th.as_tensor(-3.0))).numpy(), places=4
        )
        self.assertAlmostEqual(
            -4.0, self.atanh(self.tanh(th.as_tensor(-4.0))).numpy(), places=4
        )
        self.assertAlmostEqual(
            -5.0, self.atanh(self.tanh(th.as_tensor(-5.0))).numpy(), places=3
        )
        self.assertAlmostEqual(
            -6.0, self.atanh(self.tanh(th.as_tensor(-6.0))).numpy(), places=3
        )

        self.assertAlmostEqual(0.1, self.atanh(self.tanh(th.as_tensor(0.1))).numpy())
        self.assertAlmostEqual(0.2, self.atanh(self.tanh(th.as_tensor(0.2))).numpy())
        self.assertAlmostEqual(0.3, self.atanh(self.tanh(th.as_tensor(0.3))).numpy())
        self.assertAlmostEqual(0.4, self.atanh(self.tanh(th.as_tensor(0.4))).numpy())
        self.assertAlmostEqual(0.5, self.atanh(self.tanh(th.as_tensor(0.5))).numpy())
        self.assertAlmostEqual(0.6, self.atanh(self.tanh(th.as_tensor(0.6))).numpy())
        self.assertAlmostEqual(0.7, self.atanh(self.tanh(th.as_tensor(0.7))).numpy())
        self.assertAlmostEqual(
            0.8, self.atanh(self.tanh(th.as_tensor(0.8))).numpy(), places=6
        )
        self.assertAlmostEqual(0.9, self.atanh(self.tanh(th.as_tensor(0.9))).numpy())

    def testPWLN0D(self):
        lr = 0.01
        mlp = Sequential(
            self.pwln.app,
            nn.Linear(2, 16),
            nn.ReLU(inplace=True),
            nn.Linear(16, 16),
            nn.ReLU(inplace=True),
            nn.Linear(16, 2),
            self.pwln.inv,
        )
        optimizer = th.optim.Adam(mlp.parameters(), lr=lr)
        mse = MSELoss()
        for _ in range(10):
            for _ in range(10):
                data = self.func(th.rand(16, 2))
                loss = mse(mlp(data), data * data * data)
                print("loss", loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            with th.no_grad():
                print(
                    0.1,
                    self.func(th.as_tensor([[0.1, 0.1]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.1, 0.1]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.2,
                    self.func(th.as_tensor([[0.2, 0.2]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.2, 0.2]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.3,
                    self.func(th.as_tensor([[0.3, 0.3]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.3, 0.3]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.4,
                    self.func(th.as_tensor([[0.4, 0.4]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.4, 0.4]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.5,
                    self.func(th.as_tensor([[0.5, 0.5]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.5, 0.5]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.6,
                    self.func(th.as_tensor([[0.6, 0.6]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.6, 0.6]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.7,
                    self.func(th.as_tensor([[0.7, 0.7]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.7, 0.7]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.8,
                    self.func(th.as_tensor([[0.8, 0.8]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.8, 0.8]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.9,
                    self.func(th.as_tensor([[0.9, 0.9]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.9, 0.9]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )

        data = self.func(th.rand(1, 2))
        self.assertAlmostEqual(
            data.detach().numpy()[0, 0],
            self.pwln.inverse(self.pwln(data)).detach().numpy()[0, 0],
            places=3,
        )

    def testPWLN1D(self):
        lr = 0.01
        mlp = Sequential(
            self.pwln.app,
            nn.Conv1d(2, 16, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv1d(16, 16, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv1d(16, 2, kernel_size=3, padding=1),
            self.pwln.inv,
        )
        optimizer = th.optim.Adam(mlp.parameters(), lr=lr)
        mse = MSELoss()
        for _ in range(10):
            for _ in range(10):
                data = self.func(th.rand(16, 2, 16))
                loss = mse(mlp(data), data * data * data)
                print("loss", loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            with th.no_grad():
                print(
                    0.1,
                    self.func(th.as_tensor([[0.1, 0.1]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.1, 0.1]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.2,
                    self.func(th.as_tensor([[0.2, 0.2]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.2, 0.2]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.3,
                    self.func(th.as_tensor([[0.3, 0.3]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.3, 0.3]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.4,
                    self.func(th.as_tensor([[0.4, 0.4]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.4, 0.4]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.5,
                    self.func(th.as_tensor([[0.5, 0.5]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.5, 0.5]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.6,
                    self.func(th.as_tensor([[0.6, 0.6]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.6, 0.6]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.7,
                    self.func(th.as_tensor([[0.7, 0.7]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.7, 0.7]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.8,
                    self.func(th.as_tensor([[0.8, 0.8]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.8, 0.8]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.9,
                    self.func(th.as_tensor([[0.9, 0.9]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.9, 0.9]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )

        data = self.func(th.rand(1, 2))
        self.assertAlmostEqual(
            data.detach().numpy()[0, 0],
            self.pwln.inverse(self.pwln(data)).detach().numpy()[0, 0],
            places=3,
        )

    def testPWLN2D(self):
        lr = 0.01
        mlp = Sequential(
            self.pwln.app,
            nn.Conv2d(2, 16, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(16, 16, kernel_size=3, padding=1),
            nn.ReLU(inplace=True),
            nn.Conv2d(16, 2, kernel_size=3, padding=1),
            self.pwln.inv,
        )
        optimizer = th.optim.Adam(mlp.parameters(), lr=lr)
        mse = MSELoss()
        for _ in range(10):
            for _ in range(10):
                data = self.func(th.rand(16, 2, 16, 16))
                loss = mse(mlp(data), data * data * data)
                print("loss", loss.item())
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

            with th.no_grad():
                print(
                    0.1,
                    self.func(th.as_tensor([[0.1, 0.1]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.1, 0.1]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.2,
                    self.func(th.as_tensor([[0.2, 0.2]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.2, 0.2]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.3,
                    self.func(th.as_tensor([[0.3, 0.3]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.3, 0.3]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.4,
                    self.func(th.as_tensor([[0.4, 0.4]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.4, 0.4]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.5,
                    self.func(th.as_tensor([[0.5, 0.5]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.5, 0.5]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.6,
                    self.func(th.as_tensor([[0.6, 0.6]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.6, 0.6]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.7,
                    self.func(th.as_tensor([[0.7, 0.7]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.7, 0.7]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.8,
                    self.func(th.as_tensor([[0.8, 0.8]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.8, 0.8]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )
                print(
                    0.9,
                    self.func(th.as_tensor([[0.9, 0.9]])).detach().numpy(),
                    self.pwln.inverse(
                        self.pwln(self.func(th.as_tensor([0.9, 0.9]).reshape(1, 2)))
                    )
                    .detach()
                    .numpy(),
                )

        data = self.func(th.rand(1, 2))
        self.assertAlmostEqual(
            data.detach().numpy()[0, 0],
            self.pwln.inverse(self.pwln(data)).detach().numpy()[0, 0],
            places=3,
        )



import torch.nn as nn


class ComplexUpsample1d(nn.Module):
    def __init__(self, size=None):
        super(ComplexUpsample1d, self).__init__()
        self.size = size
        self.upsample = nn.Upsample(size=self.size, mode="linear")

    def forward(self, input):
        return self.upsample(input.real) + 1j * self.upsample(input.imag)


class ComplexUpsample2d(nn.Module):
    def __init__(self, size=None):
        super(ComplexUpsample2d, self).__init__()
        self.size = size
        self.upsample = nn.Upsample(size=self.size, mode="bilinear")

    def forward(self, input):
        return self.upsample(input.real) + 1j * self.upsample(input.imag)


class ComplexUpsample3d(nn.Module):
    def __init__(self, size=None):
        super(ComplexUpsample3d, self).__init__()
        self.size = size
        self.upsample = nn.Upsample(size=self.size, mode="trilinear")

    def forward(self, input):
        return self.upsample(input.real) + 1j * self.upsample(input.imag)



from torch import nn as nn


def conv7x7(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """7x7 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=7,
        stride=stride,
        padding=3,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv5x5(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """5x5 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=5,
        stride=stride,
        padding=2,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv3x3(in_planes, out_planes, stride=1, groups=1, dilation=1):
    """3x3 convolution with padding"""
    return nn.Conv2d(
        in_planes,
        out_planes,
        kernel_size=3,
        stride=stride,
        padding=1,
        groups=groups,
        bias=False,
        dilation=dilation,
    )


def conv1x1(in_planes, out_planes, stride=1):
    """1x1 convolution"""
    return nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=stride, bias=False)


class DepthwiseSeparableConv1d(nn.Module):
    def __init__(
        self,
        in_channels,
        output_channels,
        kernel_size=1,
        padding=0,
        stride=1,
        dilation=1,
        groups=1,
        bias=True,
        kernels_per_layer=1,
    ):
        super(DepthwiseSeparableConv1d, self).__init__()
        # In Tensorflow DepthwiseConv2D has depth_multiplier instead of kernels_per_layer
        self.depthwise = nn.Conv1d(
            in_channels,
            in_channels * kernels_per_layer,
            kernel_size,
            padding=padding,
            stride=stride,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise = nn.Conv1d(
            in_channels * kernels_per_layer,
            output_channels,
            1,
            padding=0,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class DepthwiseSeparableConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        output_channels,
        kernel_size=1,
        padding=0,
        stride=1,
        dilation=1,
        groups=1,
        bias=True,
        kernels_per_layer=1,
    ):
        super(DepthwiseSeparableConv2d, self).__init__()
        # In Tensorflow DepthwiseConv2D has depth_multiplier instead of kernels_per_layer
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels * kernels_per_layer,
            kernel_size,
            padding=padding,
            stride=stride,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise = nn.Conv2d(
            in_channels * kernels_per_layer,
            output_channels,
            1,
            padding=0,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class DepthwiseSeparableConv3d(nn.Module):
    def __init__(
        self,
        in_channels,
        output_channels,
        kernel_size=1,
        padding=0,
        stride=1,
        dilation=1,
        groups=1,
        bias=True,
        kernels_per_layer=1,
    ):
        super(DepthwiseSeparableConv3d, self).__init__()
        self.depthwise = nn.Conv3d(
            in_channels,
            in_channels * kernels_per_layer,
            kernel_size,
            padding=padding,
            stride=stride,
            dilation=dilation,
            groups=in_channels,
            bias=bias,
        )
        self.pointwise = nn.Conv3d(
            in_channels * kernels_per_layer,
            output_channels,
            1,
            padding=0,
            stride=stride,
            dilation=dilation,
            groups=groups,
            bias=bias,
        )

    def forward(self, x):
        x = self.depthwise(x)
        x = self.pointwise(x)
        return x


class ComplexConv1d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super(ComplexConv1d, self).__init__()
        self.rconv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )
        self.iconv = nn.Conv1d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )

    def forward(self, input):
        return (
            self.rconv(input.real)
            - self.iconv(input.imag)
            + 1j * (self.rconv(input.imag) + self.iconv(input.real))
        )


class ComplexConv2d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super(ComplexConv2d, self).__init__()
        self.rconv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )
        self.iconv = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )

    def forward(self, input):
        return (
            self.rconv(input.real)
            - self.iconv(input.imag)
            + 1j * (self.rconv(input.imag) + self.iconv(input.real))
        )


class ComplexConv3d(nn.Module):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=3,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=True,
    ):
        super(ComplexConv3d, self).__init__()
        self.rconv = nn.Conv3d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )
        self.iconv = nn.Conv3d(
            in_channels,
            out_channels,
            kernel_size,
            stride,
            padding,
            dilation,
            groups,
            bias,
        )

    def forward(self, input):
        return (
            self.rconv(input.real)
            - self.iconv(input.imag)
            + 1j * (self.rconv(input.imag) + self.iconv(input.real))
        )



import torch
from .misc import (
    _scaled_dot_product,
    _convert_to_tensor,
    _is_finite,
    _select_initial_step,
    _handle_unused_kwargs,
)
from .solvers import AdaptiveStepsizeODESolver
from .rk_common import _RungeKuttaState, _ButcherTableau, _runge_kutta_step

# Parameters from Tsitouras (2011).
_TSITOURAS_TABLEAU = _ButcherTableau(
    alpha=[0.161, 0.327, 0.9, 0.9800255409045097, 1.0, 1.0],
    beta=[
        [0.161],
        [-0.008480655492357, 0.3354806554923570],
        [2.897153057105494, -6.359448489975075, 4.362295432869581],
        [
            5.32586482843925895,
            -11.74888356406283,
            7.495539342889836,
            -0.09249506636175525,
        ],
        [
            5.86145544294642038,
            -12.92096931784711,
            8.159367898576159,
            -0.071584973281401006,
            -0.02826905039406838,
        ],
        [
            0.09646076681806523,
            0.01,
            0.4798896504144996,
            1.379008574103742,
            -3.290069515436081,
            2.324710524099774,
        ],
    ],
    c_sol=[
        0.09646076681806523,
        0.01,
        0.4798896504144996,
        1.379008574103742,
        -3.290069515436081,
        2.324710524099774,
        0,
    ],
    c_error=[
        0.09646076681806523 - 0.001780011052226,
        0.01 - 0.000816434459657,
        0.4798896504144996 - -0.007880878010262,
        1.379008574103742 - 0.144711007173263,
        -3.290069515436081 - -0.582357165452555,
        2.324710524099774 - 0.458082105929187,
        -1 / 66,
    ],
)


def _interp_coeff_tsit5(t0, dt, eval_t):
    t = float((eval_t - t0) / dt)
    b1 = (
        -1.0530884977290216
        * t
        * (t - 1.3299890189751412)
        * (t**2 - 1.4364028541716351 * t + 0.7139816917074209)
    )
    b2 = 0.1017 * t**2 * (t**2 - 2.1966568338249754 * t + 1.2949852507374631)
    b3 = (
        2.490627285651252793
        * t**2
        * (t**2 - 2.38535645472061657 * t + 1.57803468208092486)
    )
    b4 = (
        -16.54810288924490272
        * (t - 1.21712927295533244)
        * (t - 0.61620406037800089)
        * t**2
    )
    b5 = (
        47.37952196281928122
        * (t - 1.203071208372362603)
        * (t - 0.658047292653547382)
        * t**2
    )
    b6 = -34.87065786149660974 * (t - 1.2) * (t - 0.666666666666666667) * t**2
    b7 = 2.5 * (t - 1) * (t - 0.6) * t**2
    return [b1, b2, b3, b4, b5, b6, b7]


def _interp_eval_tsit5(t0, t1, k, eval_t):
    dt = t1 - t0
    y0 = tuple(k_[0] for k_ in k)
    interp_coeff = _interp_coeff_tsit5(t0, dt, eval_t)
    y_t = tuple(
        y0_ + _scaled_dot_product(dt, interp_coeff, k_) for y0_, k_ in zip(y0, k)
    )
    return y_t


def _optimal_step_size(
    last_step, mean_error_ratio, safety=0.9, ifactor=10.0, dfactor=0.2, order=5
):
    """Calculate the optimal size for the next Runge-Kutta step."""
    if mean_error_ratio == 0:
        return last_step * ifactor
    if mean_error_ratio < 1:
        dfactor = _convert_to_tensor(
            1, dtype=torch.float64, device=mean_error_ratio.device
        )
    error_ratio = torch.sqrt(mean_error_ratio).type_as(last_step)
    exponent = torch.tensor(1 / order).type_as(last_step)
    factor = torch.max(
        1 / ifactor, torch.min(error_ratio**exponent / safety, 1 / dfactor)
    )
    return last_step / factor


def _abs_square(x):
    return torch.mul(x, x)


class Tsit5Solver(AdaptiveStepsizeODESolver):

    def __init__(
        self,
        func,
        y0,
        rtol,
        atol,
        first_step=None,
        safety=0.9,
        ifactor=10.0,
        dfactor=0.2,
        max_num_steps=2**31 - 1,
        **unused_kwargs
    ):
        _handle_unused_kwargs(self, unused_kwargs)
        del unused_kwargs

        self.func = func
        self.y0 = y0
        self.rtol = rtol
        self.atol = atol
        self.first_step = first_step
        self.safety = _convert_to_tensor(
            safety, dtype=torch.float64, device=y0[0].device
        )
        self.ifactor = _convert_to_tensor(
            ifactor, dtype=torch.float64, device=y0[0].device
        )
        self.dfactor = _convert_to_tensor(
            dfactor, dtype=torch.float64, device=y0[0].device
        )
        self.max_num_steps = _convert_to_tensor(
            max_num_steps, dtype=torch.int32, device=y0[0].device
        )

    def before_integrate(self, t):
        if self.first_step is None:
            first_step = _select_initial_step(
                self.func, t[0], self.y0, 4, self.rtol, self.atol
            ).to(t)
        else:
            first_step = _convert_to_tensor(0.01, dtype=t.dtype, device=t.device)
        self.rk_state = _RungeKuttaState(
            self.y0,
            self.func(t[0].type_as(self.y0[0]), self.y0),
            t[0],
            t[0],
            first_step,
            tuple(map(lambda x: [x] * 7, self.y0)),
        )

    def advance(self, next_t):
        """Interpolate through the next time point, integrating as necessary."""
        n_steps = 0
        while next_t > self.rk_state.t1:
            assert (
                n_steps < self.max_num_steps
            ), "max_num_steps exceeded ({}>={})".format(n_steps, self.max_num_steps)
            self.rk_state = self._adaptive_tsit5_step(self.rk_state)
            n_steps += 1
        return _interp_eval_tsit5(
            self.rk_state.t0, self.rk_state.t1, self.rk_state.interp_coeff, next_t
        )

    def _adaptive_tsit5_step(self, rk_state):
        """Take an adaptive Runge-Kutta step to integrate the ODE."""
        y0, f0, _, t0, dt, _ = rk_state
        ########################################################
        #                      Assertions                      #
        ########################################################
        assert t0 + dt > t0, "underflow in dt {}".format(dt.item())
        for y0_ in y0:
            assert _is_finite(
                torch.abs(y0_)
            ), "non-finite values in state `y`: {}".format(y0_)
        y1, f1, y1_error, k = _runge_kutta_step(
            self.func, y0, f0, t0, dt, tableau=_TSITOURAS_TABLEAU
        )

        ########################################################
        #                     Error Ratio                      #
        ########################################################
        error_tol = tuple(
            self.atol + self.rtol * torch.max(torch.abs(y0_), torch.abs(y1_))
            for y0_, y1_ in zip(y0, y1)
        )
        tensor_error_ratio = tuple(
            y1_error_ / error_tol_ for y1_error_, error_tol_ in zip(y1_error, error_tol)
        )
        sq_error_ratio = tuple(
            torch.mul(tensor_error_ratio_, tensor_error_ratio_)
            for tensor_error_ratio_ in tensor_error_ratio
        )
        mean_error_ratio = sum(
            torch.sum(sq_error_ratio_) for sq_error_ratio_ in sq_error_ratio
        ) / sum(sq_error_ratio_.numel() for sq_error_ratio_ in sq_error_ratio)
        accept_step = mean_error_ratio <= 1

        ########################################################
        #                   Update RK State                    #
        ########################################################
        y_next = y1 if accept_step else y0
        f_next = f1 if accept_step else f0
        t_next = t0 + dt if accept_step else t0
        dt_next = _optimal_step_size(
            dt, mean_error_ratio, self.safety, self.ifactor, self.dfactor
        )
        k_next = k if accept_step else self.rk_state.interp_coeff
        rk_state = _RungeKuttaState(y_next, f_next, t0, t_next, dt_next, k_next)
        return rk_state



import os
import torch as th
import leibniz as lbnz
import matplotlib

matplotlib.use("Agg")

import matplotlib.pyplot as plt  # noqa: E402

from matplotlib.animation import FuncAnimation  # noqa: E402


def binary(tensor):
    return th.where(tensor > lbnz.zero, lbnz.one, lbnz.zero)


def draw_3D_plot(x, y, fld, fig_name="test.png"):
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    fld = fld.cpu().numpy()

    fig = plt.figure()
    ax = fig.gca(projection="3d")
    c = ax.plot_surface(
        x,
        y,
        fld,
        rstride=1,
        cstride=1,
        cmap="binary",
        edgecolor="none",
        vmin=-2,
        vmax=2,
    )
    ax.set_xlim(-16, 16)
    ax.set_ylim(-16, 16)
    ax.set_zlim(-2, 2)

    fig.colorbar(c)
    plt.savefig(os.path.join("results", fig_name))
    plt.close()


def draw_animation(sequence, x, y, gif_name="test.gif", frames=51, mode="3d"):
    gif_name = os.path.join("results", gif_name)
    x = x.cpu().numpy()
    y = y.cpu().numpy()
    sequence = [item.cpu().numpy() for item in sequence]

    fig = plt.figure()

    def update(n):
        print("drawing %2dth image in animation" % n)
        plt.clf()
        ax = fig.gca(projection="3d")
        c = ax.plot_surface(
            x,
            y,
            sequence[n],
            rstride=1,
            cstride=1,
            cmap="binary",
            edgecolor="none",
            vmin=-2,
            vmax=2,
        )
        ax.set_xlim(-16, 16)
        ax.set_ylim(-16, 16)
        ax.set_zlim(-2, 2)
        fig.colorbar(c)

    ani = FuncAnimation(
        fig,
        update,
        frames=frames,
        interval=1000.0 / 4,
        save_count=100,
        repeat_delay=1000,
        repeat=True,
    )

    # save it into gif (need imagemagick)
    ani.save(gif_name[:-4] + mode + ".gif", writer="imagemagick")
    plt.close()


def test_Gaussian_rotation():
    fld = th.exp(-(((lbnz.x - 1) * 2) ** 2 + ((lbnz.y - 1) * 2) ** 2))  # Gaussian

    r = th.sqrt(lbnz.x**2 + lbnz.y**2)
    theta = th.atan2(lbnz.y, lbnz.x)
    wind = -r * th.sin(theta), r * th.cos(theta), 0

    sequence, test = [fld], fld
    for t in range(1, t_iter + 1):
        adv_increment = lbnz.adv(wind, test)
        if t == 1:
            test = sequence[t - 1] - adv_increment * dt
        else:
            test = sequence[t - 2] - adv_increment * 2 * dt
        sequence.append(test)

    sequence = [sequence[i].squeeze()[..., lbnz.H // 2] for i in range(0, t_iter, 10)]
    draw_animation(
        sequence,
        lbnz.x.squeeze()[..., 0],
        lbnz.y.squeeze()[..., 0],
        "simple_rotation1.gif",
        mode="3d",
        frames=len(sequence),
    )


def test_cube_rotation():
    fld = (
        binary((lbnz.x - 2) * (3 - lbnz.x))
        * binary((lbnz.y - 2) * (3 - lbnz.y))
        * binary((lbnz.z - 1) * (6 - lbnz.z))
    )

    r = th.sqrt(lbnz.x**2 + lbnz.y**2)
    theta = th.atan2(lbnz.y, lbnz.x)
    wind = -r * th.sin(theta), r * th.cos(theta), 0

    sequence, test = [fld], fld
    for t in range(1, t_iter + 1):
        adv_increment = lbnz.adv(wind, test)
        if t == 1:
            test = sequence[t - 1] - adv_increment * dt
        else:
            test = sequence[t - 2] - adv_increment * 2 * dt
        sequence.append(test)

    sequence = [sequence[i].squeeze()[..., lbnz.H // 2] for i in range(0, t_iter, 10)]
    draw_animation(
        sequence,
        lbnz.x.squeeze()[..., 0],
        lbnz.y.squeeze()[..., 0],
        "simple_rotation2.gif",
        mode="3d",
        frames=len(sequence),
    )


if __name__ == "__main__":
    from leibniz.core3d.gridsys.regular3 import RegularGrid

    lbnz.bind(
        RegularGrid(
            basis="x,y,z",
            W=151,
            L=151,
            H=15,
            east=16.0,
            west=-16.0,
            north=16.0,
            south=-16.0,
            upper=6.0,
            lower=1.0,
        )
    )
    lbnz.use("x,y,z")

    dt = 1 / 120
    t_iter = 1000

    # test_Gaussian_rotation()
    test_cube_rotation()



from torch import nn as nn


class ComplexAvgPool1d(nn.Module):
    def __init__(self):
        super(ComplexAvgPool1d, self).__init__()
        self.pool = nn.AdaptiveAvgPool1d(1)

    def forward(self, input):
        return self.pool(input.real) + 1j * self.pool(input.imag)


class ComplexAvgPool2d(nn.Module):
    def __init__(self):
        super(ComplexAvgPool2d, self).__init__()
        self.pool = nn.AdaptiveAvgPool2d(1)

    def forward(self, input):
        return self.pool(input.real) + 1j * self.pool(input.imag)


class ComplexAvgPool3d(nn.Module):
    def __init__(self):
        super(ComplexAvgPool3d, self).__init__()
        self.pool = nn.AdaptiveAvgPool3d(1)

    def forward(self, input):
        return self.pool(input.real) + 1j * self.pool(input.imag)



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import torch as th

import leibniz as lbnz

from torch import Tensor
from leibniz.core3d.gridsys.regular3 import RegularGrid


class TestOpsPhiThetaR(unittest.TestCase):

    def setUp(self):
        lbnz.bind(
            RegularGrid(
                basis="lng,lat,alt",
                L=51,
                W=51,
                H=51,
                east=119.0,
                west=114.0,
                north=42.3,
                south=37.3,
                upper=16000.0,
                lower=0.0,
            ),
            r0=6371000,
        )
        lbnz.use("theta,phi,r")
        lbnz.use("thetaphir")

    def tearDown(self):
        lbnz.clear()

    def magnitude(self, var):
        # return magnitude of variable to test accuracy
        if th.is_tensor(var):
            var = var.cpu().numpy()
        if isinstance(var, np.ndarray):
            return 10 ** int(np.log10(np.abs(var.max()) + 1))
        else:
            return 10 ** int(np.log10(np.abs(var) + 1))

    def assertAlmostEqualWithMagnitude(self, excepted, test, magnitude=6):
        if (
            isinstance(excepted, Tensor)
            and th.abs(excepted).max().cpu().numpy() == 0.0
            or isinstance(excepted, float)
            and excepted == 0.0
        ):
            return self.assertAlmostEqual(0.0, test.max().cpu().numpy(), magnitude)
        else:
            mag = self.magnitude(excepted)
            err = th.abs(excepted - test) / mag
            return self.assertAlmostEqual(0.0, err.max().cpu().numpy(), magnitude)

    def test_grad(self):
        dLth, dLph, dLr = lbnz.dL

        g0, g1, g2 = lbnz.grad(lbnz.lng)

        self.assertAlmostEqualWithMagnitude(5.0, g0 * dLth * (lbnz.L - 1), 6)
        self.assertAlmostEqualWithMagnitude(0.0, g1, 6)
        self.assertAlmostEqualWithMagnitude(0.0, g2, 6)

        g0, g1, g2 = lbnz.grad(lbnz.lat)

        self.assertAlmostEqualWithMagnitude(0.0, g0, 6)
        self.assertAlmostEqualWithMagnitude(5.0, g1 * dLph * (lbnz.W - 1), 6)
        self.assertAlmostEqualWithMagnitude(0.0, g2, 6)

        g0, g1, g2 = lbnz.grad(lbnz.alt)

        self.assertAlmostEqualWithMagnitude(0.0, g0, 6)
        self.assertAlmostEqualWithMagnitude(0.0, g1, 6)
        self.assertAlmostEqualWithMagnitude(16000.0, g2 * dLr * (lbnz.H - 1), 6)

    def test_zero_curl(self):
        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.lat))

        self.assertAlmostEqualWithMagnitude(0, g0, 6)
        self.assertAlmostEqualWithMagnitude(0, g1, 6)
        self.assertAlmostEqualWithMagnitude(0, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.lng))

        self.assertAlmostEqualWithMagnitude(0, g0, 6)
        self.assertAlmostEqualWithMagnitude(0, g1, 6)
        self.assertAlmostEqualWithMagnitude(0, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.alt))

        self.assertAlmostEqualWithMagnitude(0, g0, 6)
        self.assertAlmostEqualWithMagnitude(0, g1, 6)
        self.assertAlmostEqualWithMagnitude(0, g2, 6)

        g0, g1, g2 = lbnz.curl(lbnz.grad(lbnz.alt * lbnz.alt + lbnz.lng * lbnz.lat))

        self.assertAlmostEqualWithMagnitude(0, g0, 6)
        self.assertAlmostEqualWithMagnitude(0, g1, 6)
        self.assertAlmostEqualWithMagnitude(0, g2, 6)

    def test_zero_div(self):
        g = lbnz.div(lbnz.curl(lbnz.thetaphir.phi))
        self.assertAlmostEqualWithMagnitude(0, g, 6)

        g = lbnz.div(lbnz.curl(lbnz.thetaphir.theta))
        self.assertAlmostEqualWithMagnitude(0, g, 6)

        g = lbnz.div(lbnz.curl(lbnz.thetaphir.r))
        self.assertAlmostEqualWithMagnitude(0, g, 6)

        phx, phy, phz = lbnz.thetaphir.phi
        thx, thy, thz = lbnz.thetaphir.theta
        rx, ry, rz = lbnz.thetaphir.r
        g = lbnz.div(
            lbnz.curl((rx * rx + phx * thx, ry * ry + phy * thy, rz * rz + phz * thz))
        )
        self.assertAlmostEqualWithMagnitude(0, g, 6)

    def test_div1(self):
        # F = (-theta, phi * theta, r);
        # expect that div_F = theta * tan(phi) / r + phi / (r * cos(phi)) + 3

        fld = -lbnz.theta, lbnz.phi * lbnz.theta, lbnz.r

        expected = (
            lbnz.theta * th.tan(lbnz.phi) / lbnz.r
            + lbnz.phi / (lbnz.r * th.cos(lbnz.phi))
            + 3
        )

        test = lbnz.div(fld)

        self.assertAlmostEqualWithMagnitude(expected, test, 3)

    def test_div2(self):
        # phi0 = phi[:, :, W//2, L//2, H//2], theta0 = theta[:, :, W//2, L//2, H//2], r0 = r[:, :, W//2, L//2, H//2]
        # F = (-(theta - theta0), (phi - phi0) * (theta-theta0), (r-r0));
        # expect that div_F = (theta - theta0) * tan(phi) / r + (phi - phi0) / (r * cos(phi)) + 3 - 2 * r0 / r

        phi0 = lbnz.phi[:, :, lbnz.W // 2, lbnz.L // 2, lbnz.H // 2]
        theta0 = lbnz.theta[:, :, lbnz.W // 2, lbnz.L // 2, lbnz.H // 2]
        r0 = lbnz.r[:, :, lbnz.W // 2, lbnz.L // 2, lbnz.H // 2]

        Fx = -(lbnz.theta - theta0)
        Fy = (lbnz.phi - phi0) * (lbnz.theta - theta0)
        Fz = lbnz.r - r0

        test = lbnz.div((Fx, Fy, Fz))

        expected = (
            (lbnz.theta - theta0) * th.tan(lbnz.phi) / lbnz.r
            + (lbnz.phi - phi0) / (lbnz.r * th.cos(lbnz.phi))
            + 3
            - 2 * r0 / lbnz.r
        )

        self.assertAlmostEqualWithMagnitude(expected, test, 3)

    def test_curl1(self):
        # F = (-theta, phi * theta, r);
        # expect that curl_F = (-phi*theta/r, -theta/r, (theta-phi*theta*tan(phi)+1/cos(phi))/r)

        fld = -lbnz.theta, lbnz.phi * lbnz.theta, lbnz.r

        expected_ph = -lbnz.phi * lbnz.theta / lbnz.r
        expected_th = -lbnz.theta / lbnz.r
        expected_r = (
            lbnz.theta - lbnz.phi * lbnz.theta * th.tan(lbnz.phi) + 1 / th.cos(lbnz.phi)
        ) / lbnz.r

        test_ph, test_th, test_r = lbnz.curl(fld)

        self.assertAlmostEqualWithMagnitude(expected_ph, test_ph, 6)
        self.assertAlmostEqualWithMagnitude(expected_th, test_th, 6)
        self.assertAlmostEqualWithMagnitude(expected_r, test_r, 6)

    def test_curl2(self):
        # F = (0, cos(phi), 0)
        # expect that curl = (-cos(phi)/r, 0, -2*sin(phi)/r)
        fld = 0, th.cos(lbnz.phi), 0

        expected_ph = -th.cos(lbnz.phi) / lbnz.r
        expected_th = lbnz.zero
        expected_r = -2 * th.sin(lbnz.phi) / lbnz.r

        test_ph, test_th, test_r = lbnz.curl(fld)

        self.assertAlmostEqualWithMagnitude(expected_ph, test_ph, 6)
        self.assertAlmostEqualWithMagnitude(expected_th, test_th, 6)
        self.assertAlmostEqualWithMagnitude(expected_r, test_r, 6)



# -*- coding: utf-8 -*-

import numpy as np
import torch as th
import logging

from typing import Tuple
from torch import Tensor
from torch.nn.functional import conv2d, conv3d

default_device = -1
used_modules = []
used_environ = {}

logger = logging.getLogger("leibniz")

lanczosk2d = None
lanczosk3d = None


def get_device():
    return default_device


def set_device(ix):
    global default_device, lanczosk2d, lanczosk3d

    if default_device != -1 and th.cuda.is_available():
        if lanczosk2d is not None:
            lanczosk2d = lanczosk2d.cuda(device=ix)

        if lanczosk3d is not None:
            lanczosk3d = lanczosk3d.cuda(device=ix)

    if default_device == ix:
        return

    logger.info("reset device from %d to %d", default_device, ix)

    if _grid_ is not None:
        _grid_.set_device(ix)
    if _frame_ is not None:
        _frame_.set_device(ix)
    if _element_ is not None:
        _element_.set_device(ix)

    if default_device != -1 and ix == -1:
        if lanczosk2d is not None:
            lanczosk2d = lanczosk2d.cpu()

        if lanczosk3d is not None:
            lanczosk3d = lanczosk3d.cpu()

    default_device = ix

    for mname in used_modules:
        use(mname, **used_environ[mname])

    globals().update(
        {
            "L": _grid_.L,
            "W": _grid_.W,
            "H": _grid_.H,
            "boundary": lambda: _grid_.boundary(),
            "mk_zero": _grid_.mk_zero,
            "mk_one": _grid_.mk_one,
            "zero": _grid_.zero,
            "one": _grid_.one,
            "random": lambda: _grid_.random(),
            "dL": _element_.dL,
            "dS": _element_.dS,
            "dV": _element_.dV,
        }
    )


def cast(element, device=-1) -> Tensor:
    element = np.array(element, dtype=np.float64)
    tensor = th.DoubleTensor(element)
    if device != -1 and th.cuda.is_available():
        return tensor.cuda(device=device)
    else:
        return tensor


diff = None
upwindp = None
upwindm = None

_grid_ = None
_frame_ = None
_element_ = None
_ops_ = None
_v3_ = None


def clear():
    global _grid_, _frame_, _element_, _ops_, _v3_
    _grid_ = None
    _frame_ = None
    _element_ = None
    _ops_ = None
    _v3_ = None

    used_modules.clear()
    used_environ.clear()


def load(mname):
    import importlib

    if "," in mname:
        return importlib.import_module(
            "leibniz.core3d.basis.%s" % mname.replace(",", "_")
        )
    else:
        return importlib.import_module("leibniz.core3d.frame.%s" % mname)


def use(mname, **kwargs):
    if mname not in used_modules:
        used_modules.append(mname)
        used_environ[mname] = kwargs

    if "," not in mname:
        target = load(mname)
        pname = target._name_
        glbs = globals()
        pval = target._clazz_(_grid_)
        glbs.update(
            {
                pname: pval,
            }
        )
    else:
        data = _grid_.data()
        delta = _grid_.delta()

        if mname == _grid_.basis:
            target = load(mname)
            pvalues = data[:, 0:1], data[:, 1:2], data[:, 2:3]
            dvalues = delta[:, 0:1], delta[:, 1:2], delta[:, 2:3]
        else:
            origin = load(_grid_.basis)
            target = load(mname)

            tval1, tval2, tval3 = origin.transform(
                data[:, 0:1], data[:, 1:2], data[:, 2:3], **kwargs
            )
            pval1, pval2, pval3 = target.inverse(tval1, tval2, tval3, **kwargs)
            mval1, mval2, mval3 = origin.dtransform(
                data[:, 0:1],
                data[:, 1:2],
                data[:, 2:3],
                delta[:, 0:1],
                delta[:, 1:2],
                delta[:, 2:3],
                **kwargs
            )
            dval1, dval2, dval3 = target.dinverse(
                tval1, tval2, tval3, mval1, mval2, mval3, **kwargs
            )

            pvalues = pval1, pval2, pval3
            dvalues = dval1, dval2, dval3

        for ix, pname in enumerate(target._params_):
            dname = "d%s" % pname
            pval = pvalues[ix]
            dval = dvalues[ix]

            globals().update({pname: pval, dname: dval})


def bind(grid, **kwargs):
    if "device" in kwargs.keys():
        device_to = kwargs["device"]
    else:
        device_to = default_device

    logger.info("device_to: %d", device_to)

    import leibniz.core3d.vec3 as v3
    import leibniz.core3d.diffr.regular3 as d

    from leibniz.core3d.element import Elements
    from leibniz.core3d.ops import LocalOps

    global _grid_, _frame_, _element_, _ops_, _v3_, diff, upwindm, upwindp

    _grid_ = grid
    diff = d.central5
    upwindp = d.upwind_p4
    upwindm = d.upwind_m4

    globals().update(
        {
            "L": _grid_.L,
            "W": _grid_.W,
            "H": _grid_.H,
            "boundary": _grid_.boundary(),
            "zero": _grid_.zero,
            "one": _grid_.one,
            "random": _grid_.random(),
        }
    )

    use(_grid_.basis, **kwargs)

    _element_ = Elements(**kwargs)
    _ops_ = LocalOps(_grid_, _element_, diff)
    _v3_ = v3

    globals().update(
        {
            "dL": _element_.dL,
            "dS": _element_.dS,
            "dV": _element_.dV,
        }
    )

    set_device(device_to)
    d.set_device(device_to)


def dot(x: Tuple[Tensor, Tensor, Tensor], y: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.dot(x, y)


def cross(x: Tuple[Tensor, Tensor, Tensor], y: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.cross(x, y)


def norm(v: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.norm(v)


def normsq(v: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _v3_.normsq(v)


def normalize(v: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _v3_.normalize(v)


def box(
    a: Tuple[Tensor, Tensor, Tensor],
    b: Tuple[Tensor, Tensor, Tensor],
    c: Tuple[Tensor, Tensor, Tensor],
) -> Tensor:
    return _v3_.box(a, b, c)


def det(transform: Tensor) -> Tensor:
    return _v3_.det(transform)


def grad(scalar: Tensor) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.grad(scalar)


def div(vector: Tuple[Tensor, Tensor, Tensor]) -> Tensor:
    return _ops_.div(vector)


def curl(vector: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.curl(vector)


def laplacian(scalar: Tensor) -> Tensor:
    return _ops_.laplacian(scalar)


def vlaplacian(vector: Tuple[Tensor, Tensor, Tensor]) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.vlaplacian(vector)


def adv(wind: Tuple[Tensor, Tensor, Tensor], scalar: Tensor, filter=None) -> Tensor:
    return _ops_.adv(wind, scalar, filter)


def vadv(
    wind: Tuple[Tensor, Tensor, Tensor],
    vector: Tuple[Tensor, Tensor, Tensor],
    filter=None,
) -> Tuple[Tensor, Tensor, Tensor]:
    return _ops_.vadv(wind, vector, filter)


def upwind(wind: Tuple[Tensor, Tensor, Tensor], scalar: Tensor, filter=None) -> Tensor:
    return _ops_.upwind(wind, scalar, filter, upwindp=upwindp, upwindm=upwindm)


avgk = (
    cast(
        [
            [
                [
                    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                    [[1, 1, 1], [1, 1, 1], [1, 1, 1]],
                ]
            ]
        ]
    )
    / 27.0
)


def avgfilter(f):
    return conv3d(f, avgk, padding=1)


lanczosk2d = cast(
    np.sinc(
        [
            [
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [-3, -2, -1, 0, +1, +2, +3],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ]
            ]
        ]
    )
    * np.sinc(
        [
            [
                [
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [-3 / 3, -2 / 3, -1 / 3, 0, +1 / 3, +2 / 3, +3 / 3],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                    [0, 0, 0, 0, 0, 0, 0],
                ]
            ]
        ]
    )
)


lanczosk2d = lanczosk2d * lanczosk2d.transpose(-2, -1)


def lanczosfilter(f):
    return conv2d(f, lanczosk2d, padding=3)


lanczosk3d = cast(
    np.sinc(
        [
            [
                [
                    [
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [-3, -2, -1, 0, +1, +2, +3],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                    ]
                ]
            ]
        ]
    )
    * np.sinc(
        [
            [
                [
                    [
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [-3 / 3, -2 / 3, -1 / 3, 0, +1 / 3, +2 / 3, +3 / 3],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0],
                    ]
                ]
            ]
        ]
    )
)


lanczosk3d = lanczosk3d * lanczosk3d.transpose(-2, -1) * lanczosk3d.transpose(-3, -1)


def lanczos3dfilter(f):
    return conv3d(f, lanczosk3d, padding=3)



# -*- coding: utf-8 -*-

import unittest

import numpy as np
import torch as th

import leibniz as lbnz

from leibniz.core3d.diffr.regular3 import (
    central,
    sobel3,
    sobel5,
    sharr3,
    centralx,
    centraly,
    centralz,
)
from leibniz.core3d.gridsys.regular3 import RegularGrid


class TestDiff(unittest.TestCase):

    def setUp(self):
        lbnz.bind(
            RegularGrid(
                basis="lng,lat,alt",
                W=51,
                L=51,
                H=51,
                east=119.0,
                west=114.0,
                north=42.3,
                south=37.3,
                upper=16000.0,
                lower=0.0,
            )
        )

    def tearDown(self):
        lbnz.clear()

    def magnitude(self, var):
        # return magnitude of variable to test accuracy
        if th.is_tensor(var):
            var = var.cpu().numpy()
        if isinstance(var, np.ndarray):
            return 10 ** int(np.log10(np.abs(var.max()) + 1))
        else:
            return 10 ** int(np.log10(np.abs(var) + 1))

    def assertAlmostEqualWithMagnitude(self, excepted, test, magnitude=6):
        m = self.magnitude(excepted)
        return self.assertAlmostEqual(excepted / m, test / m, magnitude)

    def test_order(self):
        kx = centralx[0, 0].cpu().numpy()
        self.assertAlmostEqual(1.0, np.max(kx[2:3, :, :] - kx[0:1, :, :]))
        self.assertAlmostEqual(0, np.max(kx[1:2, :, :]))

        ky = centraly[0, 0].cpu().numpy()
        self.assertAlmostEqual(1.0, np.max(ky[:, 2:3, :] - ky[:, 0:1, :]))
        self.assertAlmostEqual(0, np.max(ky[:, 1:2, :]))

        kz = centralz[0, 0].cpu().numpy()
        self.assertAlmostEqual(1.0, np.max(kz[:, :, 2:3] - kz[:, :, 0:1]))
        self.assertAlmostEqual(0, np.max(kz[:, :, 1:2]))

    def test_central(self):
        gx, gy, gz = central(lbnz.lng)

        self.assertAlmostEqualWithMagnitude(
            5.0, gx.mean().cpu().numpy() * (lbnz.W - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = central(lbnz.lat)

        self.assertAlmostEqualWithMagnitude(0.0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            5.0, gy.mean().cpu().numpy() * (lbnz.L - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = central(lbnz.alt)

        self.assertAlmostEqualWithMagnitude(0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            16000, gz.mean().cpu().numpy() * (lbnz.H - 1), 6
        )

    def test_sobel3(self):
        gx, gy, gz = sobel3(lbnz.lng)

        self.assertAlmostEqualWithMagnitude(
            5.0, gx.mean().cpu().numpy() * (lbnz.W - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sobel3(lbnz.lat)

        self.assertAlmostEqualWithMagnitude(0.0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            5.0, gy.mean().cpu().numpy() * (lbnz.L - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sobel3(lbnz.alt)

        self.assertAlmostEqualWithMagnitude(0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            16000, gz.mean().cpu().numpy() * (lbnz.H - 1), 6
        )

    def test_sobel5(self):
        gx, gy, gz = sobel5(lbnz.lng)

        self.assertAlmostEqualWithMagnitude(
            5.0, gx.mean().cpu().numpy() * (lbnz.W - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sobel5(lbnz.lat)

        self.assertAlmostEqualWithMagnitude(0.0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            5.0, gy.mean().cpu().numpy() * (lbnz.L - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sobel5(lbnz.alt)

        self.assertAlmostEqualWithMagnitude(0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            16000, gz.mean().cpu().numpy() * (lbnz.H - 1), 6
        )

    def test_sharr3(self):
        gx, gy, gz = sharr3(lbnz.lng)

        self.assertAlmostEqualWithMagnitude(
            5.0, gx.mean().cpu().numpy() * (lbnz.W - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sharr3(lbnz.lat)

        self.assertAlmostEqualWithMagnitude(0.0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            5.0, gy.mean().cpu().numpy() * (lbnz.L - 1), 6
        )
        self.assertAlmostEqualWithMagnitude(0.0, gz.mean().cpu().numpy(), 6)

        gx, gy, gz = sharr3(lbnz.alt)

        self.assertAlmostEqualWithMagnitude(0, gx.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(0, gy.mean().cpu().numpy(), 6)
        self.assertAlmostEqualWithMagnitude(
            16000, gz.mean().cpu().numpy() * (lbnz.H - 1), 6
        )



# -*- coding: utf-8 -*-

import torch as th


class LocalOps:
    def __init__(self, grid, elem, diff):
        self.grid = grid
        self.elem = elem
        self.diff = diff

    def grad(self, scalar):
        dL1, dL2, dL3 = self.elem.dL

        g_1, g_2, g_3 = self.diff(scalar)
        g_1, g_2, g_3 = g_1 / dL1, g_2 / dL2, g_3 / dL3

        return g_1, g_2, g_3

    def div(self, F):
        F1, F2, F3 = F
        dS1, dS2, dS3 = self.elem.dS

        val_1, _, _ = self.diff(F1 * dS1)
        _, val_2, _ = self.diff(F2 * dS2)
        _, _, val_3 = self.diff(F3 * dS3)

        return (val_1 + val_2 + val_3) / self.elem.dV

    def div_(self, F):
        F1, F2, F3 = F
        dL1, dL2, dL3 = self.elem.dL

        val_1, _, _ = self.diff(F1)
        _, val_2, _ = self.diff(F2)
        _, _, val_3 = self.diff(F3)

        return val_1 / dL1 + val_2 / dL2 + val_3 / dL3

    def curl(self, F):
        F1, F2, F3 = F
        dL1, dL2, dL3 = self.elem.dL
        dS1, dS2, dS3 = self.elem.dS

        _, a, b = self.diff(F1 * dL1)
        c, _, d = self.diff(F2 * dL2)
        e, f, _ = self.diff(F3 * dL3)

        return (f - d) / dS1, (b - e) / dS2, (c - a) / dS3

    def laplacian(self, f):
        return self.div(self.grad(f))

    def vlaplacian(self, F):
        a, b, c = self.grad(self.div(F))
        d, e, f = self.curl(self.curl(F))

        return a - d, b - e, c - f

    def adv(self, wind, data, filter=None):
        u, v, w = wind
        a, b, c = self.grad(data)

        val = a * u + b * v + c * w
        if filter is None:
            return val
        else:
            return filter(val)

    def vadv(self, wind, vector, filter=None):
        u, v, w = wind
        v1, v2, v3 = vector
        a, b, c = self.grad(v1)
        d, e, f = self.grad(v2)
        g, h, i = self.grad(v3)

        val1 = a * u + b * v + c * w
        val2 = d * u + e * v + f * w
        val3 = g * u + h * v + i * w
        if filter is None:
            return val1, val2, val3
        else:
            return filter(val1), filter(val2), filter(val3)

    def upwind(self, wind, data, filter=None, upwindp=None, upwindm=None):
        # upwind scheme using speed-splitting techniques.
        # Referring to https://web.stanford.edu/group/frg/course_work/AA214B/CA-AA214B-Ch6.pdf
        # and https://en.wikipedia.org/wiki/Upwind_scheme
        # Param :
        #   points: points to be used in upwind scheme

        # grad part
        dL1, dL2, dL3 = self.elem.dL

        g_p1, g_p2, g_p3 = upwindp(data)
        g_m1, g_m2, g_m3 = upwindm(data)

        g_p1, g_p2, g_p3 = g_p1 / dL1, g_p2 / dL2, g_p3 / dL3
        g_m1, g_m2, g_m3 = g_m1 / dL1, g_m2 / dL2, g_m3 / dL3

        # adv part
        u, v, w = wind
        zero = self.grid.zero
        u_p, u_m = th.where(u > zero, u, zero), th.where(u < zero, u, zero)
        v_p, v_m = th.where(v > zero, v, zero), th.where(v < zero, v, zero)
        w_p, w_m = th.where(w > zero, w, zero), th.where(w < zero, w, zero)

        val = (
            g_p1 * u_m + g_m1 * u_p + g_p2 * v_m + g_m2 * v_p + g_p3 * w_m + g_m3 * w_p
        )
        if filter is None:
            return val
        else:
            return filter(val)



# -*- coding: utf-8 -*-

from leibniz.nn.layer.residual import Basic
from leibniz.nn.activation import Swish

from leibniz.nn.net.unet import UNet
from leibniz.nn.net.unet2 import UNet2
from leibniz.nn.net.unet3 import UNet3
from leibniz.nn.net.unetz import UNetZ
from leibniz.nn.net.resnet import ResNet
from leibniz.nn.net.resnetz import ResNetZ
from leibniz.nn.net.hyptube import HypTube, StepwiseHypTube, LeveledHypTube
from leibniz.nn.net.conv_lstm import ConvLSTM
from leibniz.nn.net.simple import Linear, Identity, SimpleCNN2d
from leibniz.nn.net.hunet import HUNet


def identical(*args, **kwargs):
    return Identity()


def linear(in_channels, out_channels, **kwargs):
    return Linear(in_channels, out_channels, **kwargs)


def cnn2d(in_channels, out_channels):
    return SimpleCNN2d(in_channels, out_channels)


def unet4(in_channels, out_channels, spatial=(256, 256), ksize_in=7, dropout=0.1):
    return UNet(
        in_channels,
        out_channels,
        block=None,
        relu=Swish(),
        attn=None,
        layers=4,
        ratio=0,
        ksize_in=ksize_in,
        dropout=dropout,
        vblks=[0, 0, 0, 0],
        hblks=[0, 0, 0, 0],
        scales=[-1, -1, -1, -1],
        factors=[1, 1, 1, 1],
        spatial=spatial,
    )


def unet8(in_channels, out_channels, spatial=(256, 256), ksize_in=7, dropout=0.1):
    return UNet(
        in_channels,
        out_channels,
        block=None,
        relu=Swish(),
        attn=None,
        layers=8,
        ratio=0,
        ksize_in=ksize_in,
        dropout=dropout,
        vblks=[0, 0, 0, 0, 0, 0, 0, 0],
        hblks=[0, 0, 0, 0, 0, 0, 0, 0],
        scales=[-1, -1, -1, -1, -1, -1, -1, -1],
        factors=[1, 1, 1, 1, 1, 1, 1, 1],
        spatial=spatial,
    )


def resunet(
    in_channels,
    out_channels,
    block=Basic,
    relu=Swish(),
    attn=None,
    layers=4,
    ratio=0,
    ksize_in=7,
    dropout_prob=0.5,
    vblks=[1, 1, 1, 1],
    hblks=[1, 1, 1, 1],
    scales=[-1, -1, -1, -1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="batch",
    enhencer=None,
    final_normalized=False,
):
    return UNet(
        in_channels,
        out_channels,
        block=block,
        relu=relu,
        attn=attn,
        layers=layers,
        ratio=ratio,
        ksize_in=ksize_in,
        dropout_prob=dropout_prob,
        vblks=vblks,
        hblks=hblks,
        scales=scales,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
        enhencer=enhencer,
        final_normalized=final_normalized,
    )


def resunet2(
    in_channels,
    out_channels,
    block=Basic,
    relu=Swish(),
    attn=None,
    layers=4,
    ratio=0,
    ksize_in=7,
    dropout_prob=0.5,
    vblks=[1, 1, 1, 1],
    hblks=[1, 1, 1, 1],
    scales=[-1, -1, -1, -1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="batch",
    enhencer=None,
    final_normalized=False,
):
    return UNet2(
        in_channels,
        out_channels,
        block=block,
        relu=relu,
        attn=attn,
        layers=layers,
        ratio=ratio,
        ksize_in=ksize_in,
        dropout_prob=dropout_prob,
        vblks=vblks,
        hblks=hblks,
        scales=scales,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
        enhencer=enhencer,
        final_normalized=final_normalized,
    )


def resunet3(
    in_channels,
    out_channels,
    block=Basic,
    relu=Swish(),
    attn=None,
    layers=4,
    ratio=0,
    ksize_in=7,
    dropout_prob=0.5,
    vblks=[1, 1, 1, 1],
    hblks=[1, 1, 1, 1],
    scales=[-1, -1, -1, -1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="batch",
    enhencer=None,
    final_normalized=False,
):
    return UNet3(
        in_channels,
        out_channels,
        block=block,
        relu=relu,
        attn=attn,
        layers=layers,
        ratio=ratio,
        ksize_in=ksize_in,
        dropout_prob=dropout_prob,
        vblks=vblks,
        hblks=hblks,
        scales=scales,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
        enhencer=enhencer,
        final_normalized=final_normalized,
    )


def resunetz(
    in_channels,
    out_channels,
    relu=Swish(),
    layers=4,
    ratio=0,
    ksize_in=7,
    vblks=[1, 1, 1, 1],
    hblks=[1, 1, 1, 1],
    scales=[-1, -1, -1, -1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="tanh",
):
    return UNetZ(
        in_channels,
        out_channels,
        relu=relu,
        layers=layers,
        ratio=ratio,
        ksize_in=ksize_in,
        vblks=vblks,
        hblks=hblks,
        scales=scales,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
    )


def resnet(
    in_channels,
    out_channels,
    block=Basic,
    relu=Swish(),
    attn=None,
    layers=4,
    ratio=0,
    dropout_prob=0.1,
    vblks=[1, 1, 1, 1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="batch",
):
    return ResNet(
        in_channels,
        out_channels,
        block=block,
        relu=relu,
        attn=attn,
        layers=layers,
        ratio=ratio,
        dropout_prob=dropout_prob,
        vblks=vblks,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
    )


def resnetz(
    in_channels,
    out_channels,
    relu=Swish(),
    layers=4,
    ratio=0,
    spatial=(256, 256),
    normalizor=None,
):
    return ResNetZ(
        in_channels,
        out_channels,
        relu=relu,
        layers=layers,
        ratio=ratio,
        spatial=spatial,
        normalizor=normalizor,
    )


def hyptub(
    in_channels,
    hidden_channels,
    out_channels,
    encoder=resunet,
    decoder=resunet,
    **kwargs
):
    return HypTube(
        in_channels,
        hidden_channels,
        out_channels,
        encoder=encoder,
        decoder=decoder,
        **kwargs
    )


def hyptub_stepwise(
    in_channels,
    hidden_channels,
    out_channels,
    steps,
    encoder=resunet,
    decoder=resunet,
    **kwargs
):
    return StepwiseHypTube(
        in_channels,
        hidden_channels,
        out_channels,
        steps,
        encoder=encoder,
        decoder=decoder,
        **kwargs
    )


def hyptub_leveled(
    in_channels,
    hidden_channels,
    out_channels,
    levels,
    encoder=resunet,
    decoder=resunet,
    **kwargs
):
    return LeveledHypTube(
        in_channels,
        hidden_channels,
        out_channels,
        levels,
        encoder=encoder,
        decoder=decoder,
        **kwargs
    )


def lstm(
    input_dim,
    hidden_dim,
    kernel_size,
    num_layers=1,
    batch_first=True,
    bias=True,
    return_all_layers=False,
):
    return ConvLSTM(
        input_dim=input_dim,
        hidden_dim=hidden_dim,
        kernel_size=kernel_size,
        num_layers=num_layers,
        batch_first=batch_first,
        bias=bias,
        return_all_layers=return_all_layers,
    )


def hunet(
    in_channels,
    out_channels,
    block=Basic,
    relu=Swish(),
    attn=None,
    layers=4,
    ratio=0,
    vblks=[1, 1, 1, 1],
    scales=[-1, -1, -1, -1],
    factors=[1, 1, 1, 1],
    spatial=(256, 256),
    normalizor="batch",
):
    return HUNet(
        in_channels,
        out_channels,
        block=block,
        relu=relu,
        attn=attn,
        layers=layers,
        ratio=ratio,
        vblks=vblks,
        scales=scales,
        factors=factors,
        spatial=spatial,
        normalizor=normalizor,
    )



import torch.nn as nn

"""
Original by ZjjConan
from https://github.com/ZjjConan/SimAM/blob/master/networks/attentions/simam_module.py
modified by Mingli Yuan (Mountain)
"""


class SimAM(nn.Module):
    def __init__(self, dim, reduction=16, kernel_size=7, conv=None, e_lambda=1e-4):
        super(SimAM, self).__init__()
        self.activaton = nn.Sigmoid()
        self.e_lambda = e_lambda

    def forward(self, x):
        sz = x.size()

        if len(sz) == 3:
            b, c, w = x.size()
            n = w - 1
            dims = [2]
        if len(sz) == 4:
            b, c, h, w = x.size()
            n = w * h - 1
            dims = [2, 3]
        if len(sz) == 5:
            b, c, l, h, w = x.size()
            n = w * h * l - 1
            dims = [2, 3, 4]

        mu = x.mean(dim=dims, keepdim=True)
        square = (x - mu) * (x - mu)
        y = (
            square / (4 * (square.sum(dim=dims, keepdim=True) / n + self.e_lambda))
            + 0.5
        )

        return x * self.activaton(y)



# -*- coding: utf-8 -*-

import unittest

import numpy as np

import leibniz as lbnz


class TestV2(unittest.TestCase):

    def setUp(self):
        self.zero = lbnz.cast(np.zeros([1, 1, 2, 2]))
        self.one = lbnz.cast(np.ones([1, 1, 2, 2]))

    def tearDown(self):
        pass



# -*- coding: utf-8 -*-

import unittest
import torch as th

from leibniz.nn.net import hyptub, resunet
from leibniz.nn.net.hyptube import StepwiseHypTube, LeveledHypTube


class TestHyptub(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test2D(self):
        hyptub(1, 1, 1, spatial=(16, 16))
        hyptub(1, 1, 1, spatial=(16, 32))
        hyptub(1, 1, 1, spatial=(32, 16))
        net = hyptub(
            1, 1, 1, spatial=(32, 16), scales=[[0, -1], [0, -1], [0, -1], [0, -1]]
        )
        net(th.rand(1, 1, 32, 16))
        net = hyptub(
            1,
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="instance",
        )
        net(th.rand(1, 1, 32, 16))
        net = hyptub(
            1,
            1,
            1,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            normalizor="layer",
        )
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))
        net = hyptub(1, 1, 1, spatial=(32, 64))
        net(th.rand(1, 1, 32, 64))
        net = hyptub(1, 1, 1, spatial=(32, 64))
        net(th.rand(1, 1, 32, 64))

    def testStepwise2D(self):
        StepwiseHypTube(1, 1, 1, 2, spatial=(16, 16), encoder=resunet, decoder=resunet)
        StepwiseHypTube(1, 1, 1, 2, spatial=(16, 32), encoder=resunet, decoder=resunet)
        StepwiseHypTube(1, 1, 1, 2, spatial=(32, 16), encoder=resunet, decoder=resunet)
        net = StepwiseHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
        )
        net(th.rand(1, 1, 32, 16))
        net = StepwiseHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
            normalizor="instance",
        )
        net(th.rand(1, 1, 32, 16))
        net = StepwiseHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
            normalizor="layer",
        )
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))
        net = StepwiseHypTube(
            1, 1, 1, 2, spatial=(32, 64), encoder=resunet, decoder=resunet
        )
        net(th.rand(1, 1, 32, 64))
        net = StepwiseHypTube(
            1, 1, 1, 2, spatial=(32, 64), encoder=resunet, decoder=resunet
        )
        net(th.rand(1, 1, 32, 64))

    def testLayered2D(self):
        LeveledHypTube(1, 1, 1, 2, spatial=(16, 16), encoder=resunet, decoder=resunet)
        LeveledHypTube(1, 1, 1, 2, spatial=(16, 32), encoder=resunet, decoder=resunet)
        LeveledHypTube(1, 1, 1, 2, spatial=(32, 16), encoder=resunet, decoder=resunet)
        net = LeveledHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
        )
        net(th.rand(1, 1, 32, 16))
        net = LeveledHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
            normalizor="instance",
        )
        net(th.rand(1, 1, 32, 16))
        net = LeveledHypTube(
            1,
            1,
            1,
            2,
            spatial=(32, 16),
            scales=[[0, -1], [0, -1], [0, -1], [0, -1]],
            encoder=resunet,
            decoder=resunet,
            normalizor="layer",
        )
        net(th.rand(1, 1, 32, 16))
        net(th.rand(2, 1, 32, 16))
        net = LeveledHypTube(
            1, 1, 1, 2, spatial=(32, 64), encoder=resunet, decoder=resunet
        )
        net(th.rand(1, 1, 32, 64))
        net = LeveledHypTube(
            1, 1, 1, 2, spatial=(32, 64), encoder=resunet, decoder=resunet
        )
        net(th.rand(1, 1, 32, 64))



import torch as th
import torch.nn as nn

from torchpwl.pwl import BasePointPWL, MonoPointPWL


class InversedMonoPointPWL(BasePointPWL):
    def __init__(self, peer):
        super(InversedMonoPointPWL, self).__init__(
            peer.num_channels,
            peer.num_breakpoints,
            num_x_points=peer.num_breakpoints + 1,
        )
        self.peer = peer

    def _reset_params(self):
        self.peer._reset_params()

    def get_y_positions(self):
        return th.sort(
            self.peer.get_x_positions(), dim=1, descending=self.peer.mono < 0
        )[0]

    def get_x_positions(self):
        return th.sort(
            self.peer.get_y_positions(), dim=1, descending=self.peer.mono < 0
        )[0]


class PWLNormalizorApp(nn.Module):
    def __init__(self, num_channels, num_breakpoints, monotonicity=1):
        super(PWLNormalizorApp, self).__init__()
        self.pwl = MonoPointPWL(
            num_channels, num_breakpoints, monotonicity=monotonicity
        )
        self.pwl.mono = monotonicity

    def forward(self, x):
        sz1 = list(x.size())
        l, c = len(sz1), sz1[1]
        shift = [0] + list(range(2, l)) + [1]
        unshift = [0, l - 1] + list(range(1, l - 1))
        sz2 = [sz1[0]] + list([sz1[i] for i in range(2, l)]) + [sz1[1]]
        x = x.permute(*shift).reshape(-1, c)
        return self.pwl(x).reshape(*sz2).permute(*unshift).reshape(*sz1)


class PWLNormalizorInv(nn.Module):
    def __init__(self, app):
        super(PWLNormalizorInv, self).__init__()
        self.pwl = InversedMonoPointPWL(app.pwl)

    def forward(self, x):
        sz1 = list(x.size())
        l, c = len(sz1), sz1[1]
        shift = [0] + list(range(2, l)) + [1]
        unshift = [0, l - 1] + list(range(1, l - 1))
        sz2 = [sz1[0]] + list([sz1[i] for i in range(2, l)]) + [sz1[1]]
        x = x.permute(*shift).reshape(-1, c)
        return self.pwl(x).reshape(*sz2).permute(*unshift).reshape(*sz1)


class PWLNormalizor(nn.Module):
    def __init__(
        self, num_channels, num_breakpoints, monotonicity=1, std=1.0, mean=0.0
    ):
        super(PWLNormalizor, self).__init__()
        self.app = PWLNormalizorApp(
            num_channels, num_breakpoints, monotonicity=monotonicity
        )
        self.inv = PWLNormalizorInv(self.app)
        self.mean = mean
        self.std = std

    def forward(self, x):
        return self.app((x - self.mean) / self.std)

    def inverse(self, x):
        return self.inv(x) * self.std + self.mean


class ComplexBatchNorm1d(nn.Module):
    def __init__(
        self,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
    ):
        super(ComplexBatchNorm1d, self).__init__()
        self.rbn = nn.BatchNorm1d(
            num_features, eps, momentum, affine, track_running_stats
        )
        self.ibn = nn.BatchNorm1d(
            num_features, eps, momentum, affine, track_running_stats
        )

    def forward(self, input):
        return self.rbn(input.real) + 1j * self.ibn(input.imag)


class ComplexBatchNorm2d(nn.Module):
    def __init__(
        self,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
    ):
        super(ComplexBatchNorm2d, self).__init__()
        self.rbn = nn.BatchNorm2d(
            num_features, eps, momentum, affine, track_running_stats
        )
        self.ibn = nn.BatchNorm2d(
            num_features, eps, momentum, affine, track_running_stats
        )

    def forward(self, input):
        return self.rbn(input.real) + 1j * self.ibn(input.imag)


class ComplexBatchNorm3d(nn.Module):
    def __init__(
        self,
        num_features,
        eps=1e-5,
        momentum=0.1,
        affine=True,
        track_running_stats=True,
    ):
        super(ComplexBatchNorm3d, self).__init__()
        self.rbn = nn.BatchNorm3d(
            num_features, eps, momentum, affine, track_running_stats
        )
        self.ibn = nn.BatchNorm3d(
            num_features, eps, momentum, affine, track_running_stats
        )

    def forward(self, input):
        return self.rbn(input.real) + 1j * self.ibn(input.imag)
